 REM Sic.bas v8.7a r6.0 - Symbolic Instruction Code; module 1 of 3.
 REM The public domain DOS programming interpreter.

 ' get standard include declarations
 REM $INCLUDE: 'SIC.INC'

 ' declare startup error routine
 ON ERROR GOTO Error.Routine2

 ' program variables
 REDIM Commands(1 TO MaxCommands) AS STRING
 REDIM Statements(1 TO MaxStatements + 3) AS STRING

 ' program code
 REDIM GosubReturn(1 TO 10) AS INTEGER
 REDIM Program(0 TO Max.Lines) AS STRING
 REDIM LineBreak(1 TO Max.Lines) AS INTEGER

 ' variable break arrays
 REDIM VariableBreak(1 TO 26) AS INTEGER
 REDIM VariableValue(1 TO 26) AS DOUBLE
 REDIM VariableBreak2(1 TO 26) AS INTEGER
 REDIM VariableValue2(1 TO 26) AS STRING
 REDIM VariableBreak3(0 TO 26, 0 TO Max.Arrays) AS INTEGER
 REDIM VariableValue3(0 TO 26, 0 TO Max.Arrays) AS DOUBLE

 ' program data
 REDIM Arrays(0 TO 26, 0 TO Max.Arrays) AS DOUBLE
 REDIM Definitions(0 TO Max.Functions) AS STRING
 REDIM Strngs(0 TO 26) AS STRING
 REDIM Variables(0 TO 26) AS DOUBLE

 ' file areas
 REDIM Field.Array(1 TO 255) AS STRING
 REDIM File.Fields(1 TO 255) AS STRING

 Redim FileTraps(1 To 255) As Integer
 Redim FileRecords(1 To 128) As Long

 ' increase stack for recursion
 STACK 16384

 ' initialize common variables
 Max.Gosubs = 10
 Program.Name = None
 Quote = CHR$(34)
 SaveOnExit = True
 StdinInput1 = False
 StdinInput2 = Nul
 StdinPrompt = True
 Token.List = " -+*/\^()[]{}<>=|&!%~?:#@`;,'" + Quote
 White.Space = CHR$(32) + CHR$(9) ' can be changed

 ' initialize default screen variables
 LastScreen1 = 0
 LastScreen2 = 1
 LastScreen3 = 0
 LastScreen4 = 0

 ' initialize local screen variables
 ScreenMode = 0
 ScreenHeight = 25
 ScreenWidth = 80

 ' initialize multi-dimensional command history arrays
 Redim History2(1 To 3, 1 To Max.History) As String*256
 Redim History.Count2(1 To 3) As Integer
 Redim EndOfHistoryFlag(0 To 3) As Integer

 ' reset multi-dimensional command history arrays
 For VarQ1=1 To 3
    EndOfHistoryFlag(VarQ1) = 0
    History.Count2(VarQ1)=0
    For VarQ2=1 To Max.History
       History2(VarQ1,VarQ2)=Nul
    Next
 Next

 ' check windows
 Windows.Detected = True
 InregsX.AX = &H160A
 CALL InterruptX(&H2F, InregsX, OutregsX)
 IF OutregsX.AX > 0 THEN
    InregsX.AX = &H4A33
    CALL InterruptX(&H2F, InregsX, OutregsX)
    IF OutregsX.AX = 0 THEN
       Windows.Detected = 0 ' DOS 7.00
    END IF
 END IF

 ' store current drive/directory
 CurrentDirectory2 = CURDIR$

 ' read error strings
 CALL Read.Config(File.Found)

 ' reseed randomizer
 RANDOMIZE TIMER

 ' get environment variables
 Var$ = ENVIRON$("SICPAGELENGTH")
 IF LEN(Var$) THEN
    Temp# = INT(VAL(Var$))
    IF Temp# > 0# AND Temp# <= 32767# THEN
       PageLength = CINT(Temp#)
    END IF
 END IF
 Var$ = Environ$("SICSTATUSLINE")
 If Len(Var$) Then
    If UCase$(Var$) = "OFF" Then
       Debug.Mode = -2
    End If
 End If
 Var$ = ENVIRON$("SICSTDININPUT")
 IF LEN(Var$) THEN
    StdinInput1 = INT(VAL(Var$))
    StdinInput2 = Var$
 END IF
 Var$ = ENVIRON$("SICSTDINPROMPT")
 IF LEN(Var$) THEN
    StdinPrompt = False
 END IF
 Var$ = ENVIRON$("SICSAVEONEXIT")
 IF LEN(Var$) THEN
    SaveOnExit = False
 END IF
 Prompt1 = ">"
 Var$ = ENVIRON$("SICPROMPT1")
 IF LEN(Var$) THEN
    Prompt1 = Var$
 END IF
 Prompt2 = ":"
 Var$ = ENVIRON$("SICPROMPT2")
 IF LEN(Var$) THEN
    Prompt2 = Var$
 END IF
 Prompt3 = "?"
 Var$ = ENVIRON$("SICPROMPT3")
 IF LEN(Var$) THEN
    Prompt3 = Var$
 END IF
 Var$ = ENVIRON$("SICIGNORE")
 IF LEN(Var$) THEN
    Boot.Error = True
 END IF

 ' read in commands
 RESTORE Command.Data
 FOR Temp = 1 TO MaxCommands
    READ Commands(Temp)
 NEXT

 ' read in statements
 RESTORE Program.Data
 FOR Temp = 1 TO MaxStatements + 3
    READ Statements(Temp)
 NEXT

 ' store basic dta
 InregsX.AX = &H2F00
 CALL InterruptX(&H21, InregsX, OutregsX)
 BASIC.DTA.SEG = OutregsX.ES
 BASIC.DTA.OFF = OutregsX.BX

 ' parse command line
 Filename = COMMAND$
 IF Filename = "/?" OR Filename = "-?" THEN
    GOTO Boot.Usage
 END IF
 Var = InStr(Filename, "/Q")
 If Var Then
    Debug.Mode = -2
    Filename = Left$(Filename, Var - 1) + Mid$(Filename, Var + 2)
 End If
 Var = INSTR(Filename, "/X")
 IF Var Then
    Linux = True
    Filename = LEFT$(Filename, Var - 1) + MID$(Filename, Var + 2)
 END IF
 Var = INSTR(Filename, "/Z")
 IF Var Then
    Boot.Error = True
    Filename = LEFT$(Filename, Var - 1) + MID$(Filename, Var + 2)
 END IF
 Filename = Rtrim$(Filename)
 Filename = Ltrim$(Filename)
 IF LEFT$(FileName, 1) = "/" THEN
    Node = INT(VAL(MID$(FileName, 2, 1)))
    IF Node >= 1 AND Node <= 9 THEN
       FileName = MID$(FileName, 3)
    ELSE
       GOTO Boot.Usage
    END IF
 END IF
 Filename = Rtrim$(Filename)
 Filename = Ltrim$(Filename)
 IF LEN(Filename) THEN
    IF LEFT$(Filename, 1) = CHR$(34) THEN
       Filename = MID$(Filename, 2)
       IF RIGHT$(Filename, 1) = CHR$(34) THEN
          Filename = LEFT$(Filename, LEN(Filename) - 1)
       ELSE
          GOTO Boot.Usage
       END IF
    END IF
    IF Filename = Nul THEN
       GOTO Boot.Usage
    END IF
    IF DIR$(Filename) = Nul THEN
       GOTO Boot.Usage
    END IF
    Program.Resume = True
 END IF

 REM Increase maximum file handles
 InregsX.AX = &H6200 ' get psp segment
 CALL InterruptX(&H21, InregsX, OutregsX)
 PSP.Segment = OutregsX.BX
 DEF SEG = PSP.Segment ' point to psp segment
 ' offset to the command line
 Command.Line = &H80
 ' start of command line to store file handles
 Command.Tail = Command.Line + 20
 ' reset maximum file handles plus 5 for dos handles
 FOR Var = 1 TO 105
    POKE Command.Tail + Var - 1, &HFF
 NEXT
 ' copy first 4 file handles to our offset
 FOR Var = 1 TO 4
    File.Handle = PEEK(&H18 + Var - 1)
    POKE Command.Tail + Var - 1, File.Handle
 NEXT
 POKE &H32, 105 ' store maximum file handles
 POKE &H34, Command.Tail ' store offset to new file handles
 DEF SEG

 ' declare tempfilenames
 CALL KillTemp(0)
 CALL MakeTemp(0)
 CALL KillTemp(Node)
 CALL MakeTemp(Node)
 IF Node = 0 THEN
    DRX$ = "C:\TEMP\"
 ELSE
    DRX$ = "C:\TEMP" + LTRIM$(STR$(Node)) + "\"
 END IF

 ' declare main input loop error routine
 ON ERROR GOTO Error.Routine

 ' make operating system string
 CALL OS

 ' program.resume variable:
 '   -1 = command line program
 '    0 = immediate interpreter
 '    1 = standard input program

 ' run program from command line
 IF LEN(Filename) THEN
    Filename=UCASE$(Filename)
    CALL Read.Rows(Max.Row)
    CALL StatusLine
    StdinEnabled = -1
    Program.Resume = True
    CALL New.Program
    CALL Read.Program
    Run.Type = False
    CALL Run.Program(False)
    CALL Stop.Program
 END IF

 ' init screen size
 CALL Read.Rows(Max.Row)
 CALL Save.Screen2(0)
 CALL StatusLine

 ' run program from stdin
 CALL Read.Stdin

 ' display logon banner
 COLOR White, Black
 PRINT "SIC v" + Version + " r" + Release + ": ";
 PRINT "The Symbolic Instruction Code Interpreter."
 PRINT "Type 'Quit' to return to system."

 ' check read.config successful
 IF File.Found = False THEN
    COLOR Yellow, Black
    PRINT "File '"+ErrorList+"' not found. Error list truncated."
 END IF
 COLOR White, Black

 ' declare main input loop error routine
 ON ERROR GOTO Error.Routine

' enter the Sic engine
Error.Resume:
 ' check which function started Sic
 IF Program.Resume THEN
    CALL End.Program
    CALL Stop.Program
 END IF
 IF FileNumber THEN
    CLOSE #FileNumber
    FileNumber = False
 END IF

 ' declare main input loop error routine
 ON ERROR GOTO Error.Routine

 ' enter the Sic programming interface
 DO
    ' get input
    Var = ClearBreak
    Valid.Command = False
    Visible = 1
    COLOR Yellow, Black
    LOCATE , 1, 1
    PRINT Prompt1;
    LOCATE , , 1
    Color 7,0
    Temp0 = 1
    Out2 = RTRIM$(KeyboardLine2$)
    PRINT
    IF BreakIS THEN
       COLOR Red, Black
       PRINT "*break*":SOUND 900,6
       SLEEP 1
    END IF
    IF Out2 = Nul THEN
       Valid.Command = True
    ELSE
       ' check preceding line number
       CALL Enter.Program.Line
       IF Assign THEN
          Valid.Command = True
       ELSE

          ' get environment variable
          Param2$ = Nul
          Var$ = ENVIRON$("SICCMDLINE")
          IF LEN(Var$) THEN
             Param2$ = Var$
          END IF

          ' store parameter
          Param$ = Nul
          Out3 = Out2
          Out2 = UCASE$(Out2)
          Parameter = INSTR(Out2, " ")
          IF Parameter THEN
             Param$ = MID$(Out2, Parameter + 1)
             Param2$ = MID$(Out3, Parameter + 1)
             Out2 = LEFT$(Out2, Parameter - 1)
          END IF

          ' compare to command list
          Valid.Command = False
          FOR Command.Number = 1 TO MaxCommands
             IF Out2 = Commands(Command.Number) OR _
             Out2 = LEFT$(Commands(Command.Number), 3) THEN

                ' check command uses currently loaded program
                SELECT CASE Command.Number
                CASE 1, 4, 6, 8, 9, 11, 12, 13, 15, 18
                IF Program.Name = None THEN
                   COLOR White, Black
                   PRINT "No program loaded."
                   Valid.Command = True
                   EXIT FOR
                END IF
                END SELECT

                ' check command has no parameter to specify
                SELECT CASE Command.Number
                CASE 1, 3, 8, 10, 15, 16, 18, 20, 21, 22
                   IF LEN(Param$) THEN
                      EXIT FOR
                   END IF
                END SELECT

                ' process command
                Valid.Command = True
                SELECT CASE Command.Number
                CASE 1 ' analyze
                   CALL Prepare.Program
                   CALL Analyze.Program(-1, VarX, Var1$)
                   CALL New.Program
                   Filename = Prepare.Filename
                   CALL Read.Program
                   KILL Prepare.Filename
                CASE 2 ' files
                   CALL List.Files(Param$, True)
                CASE 3 ' help
                   CALL List.Help
                CASE 4 ' indent
                   CALL Indent.Program(INT(VAL(Param$)))
                CASE 5 ' kill
                   CALL Kill.Program(Param$)
                CASE 6 ' list
                   Start.Line = 1
                   CALL Count.Lines(Last.Line)
                   Stop.Line = Last.Line
                   IF LEN(Param$) THEN
                      Comma = INSTR(Param$, ",")
                      IF Comma THEN
                         Start.Line = INT(VAL(LEFT$(Param$, Comma - 1)))
                         Stop.Line = INT(VAL(MID$(Param$, Comma + 1)))
                      ELSE
                         Start.Line = INT(VAL(Param$))
                         Stop.Line = Last.Line
                      END IF
                   END IF
                   CALL List.Program(Start.Line, Stop.Line, False)
                CASE 7 ' load
                   CALL Load.Program(Param$)
                CASE 8 ' new
                   IF SaveOnExit THEN
                      CALL Save.Current
                   END IF
                   Last.Search.Line = False
                   CALL New.Program
                   Program.Name = None
                   COLOR White, Black
                   PRINT "Program cleared."
                CASE 9 ' print
                   Start.Line = 1
                   CALL Count.Lines(Last.Line)
                   Stop.Line = Last.Line
                   IF LEN(Param$) THEN
                      Comma = INSTR(Param$, ",")
                      IF Comma THEN
                         Start.Line = INT(VAL(LEFT$(Param$, Comma - 1)))
                         Stop.Line = INT(VAL(MID$(Param$, Comma + 1)))
                      ELSE
                         Start.Line = INT(VAL(Param$))
                         Stop.Line = Last.Line
                      END IF
                   END IF
                   CALL List.Program(Start.Line, Stop.Line, True)
                CASE 10 ' quit
                   CALL Quit.Program
                   END
                CASE 11 ' renumber
                   Start.Line = False
                   Increment.Value = False
                   IF LEN(Param$) THEN
                      Comma = INSTR(Param$, ",")
                      IF Comma THEN
                         Start.Line = INT(VAL(LEFT$(Param$, Comma - 1)))
                         Increment.Value = INT(VAL(MID$(Param$, Comma + 1)))
                      END IF
                   END IF
                   CALL Renumber.Program(Start.Line, Increment.Value)
                CASE 12 ' run
                   CALL Analyze.Program(0, VarX, Var$)
                   If VarX Then
                      Color White
                      Print "Program analyze error line:" + Str$(Program.Line) + ": " + Var$ + "."
                      Run.Line = Max.Lines + 1
                   Else
                      CommandLine = Param2$
                      IF LEFT$(UCASE$(Param2$), 5) = "LINE=" THEN
                         Start.Number = INT(VAL(MID$(Param2$, 6)))
                      ELSE
                         Start.Number = False
                      END IF
                      COLOR White, Black
                      PRINT "Starting program: " + Program.Name
                      Run.Type=False
                      CALL Run.Program(Start.Number)
                      COLOR White, Black
                      IF POS(0) > 1 THEN
                         PRINT
                      END IF
                      PRINT "Program run ended."
                   Endif
                CASE 13 ' save
                   CALL Save.Program(Param$)
                CASE 14 ' whatis
                   CALL Count.Lines(Last.Line)
                   CALL Whatis.Command(Param$)
                CASE 15 ' continue
                   Run.Type=-2
                   CALL Count.Lines2(V%)
                   IF V% = 1 THEN
                      COLOR White, Black
                      PRINT "Continuing program: " + Program.Name
                      Run.Line = 1
                      CALL Run.Program(0)
                   ELSE
                      IF Run.Line > Max.Lines THEN
                         COLOR White, Black
                         PRINT "Can't continue."
                      ELSE
                         CALL Count.Lines(V%)
                         IF Run.Line >= V% THEN
                            COLOR White, Black
                            PRINT "Can't continue."
                         ELSE
                            CALL Analyze.Program(0, VarX, Var$)
                            If VarX Then
                               Color White
                               Print "Program analyze error line:" + Str$(Program.Line) + ": " + Var$ + "."
                               Run.Line = Max.Lines + 1
                            Else
                               COLOR White, Black
                               PRINT "Continuing program: " + Program.Name
                               CALL Run.Program(0)
                           Endif
                         Endif
                      END IF
                   END IF
                CASE 16 ' debug
                   CALL Debug.Command
                CASE 17 ' shell
                   CALL Shell.Program2(Param$)
                CASE 18
                   CALL Search.Program
                CASE 19
                   CALL Set.Prompts
                CASE 20
                   Color White, Black
                   PRINT TIME$
                CASE 21
                   Color White, Black
                   PRINT DATE$
                CASE 22
                   Color White, Black
                   PRINT DATE$+" "+TIME$
                CASE 23 ' dirs
                   CALL List.Files(Param$, False)
                CASE 24 ' drives
                   CALL List.Drives(Param$, False)
                END SELECT
                EXIT FOR
             END IF
          NEXT
       END IF
    END IF
    IF Valid.Command = False THEN
       COLOR White, Black
       PRINT "Type 'Help' for information."
    END IF
 LOOP
 END

 ' display boot usage and quit
Boot.Usage:
 CALL Read.Rows(Max.Row)
 COLOR White, Black
 PRINT "SIC v" + Version + " r" + Release + " Usage:"
 COLOR Yellow, Black
 PRINT "SIC [/n] [/X] [/Z]"
 COLOR Green, Black
 PRINT "  Starts SIC"
 COLOR Yellow, Black
 PRINT "SIC [/n] [/X] [/Z] <program name>"
 COLOR Green, Black
 PRINT "  Starts SIC and runs program."
 COLOR Yellow, Black
 PRINT "TYPE <program name> | SIC [/n] [/X] [/Z]"
 COLOR Green, Black
 PRINT "  Redirects program to SIC."
 COLOR White, Black
 PRINT "(optional /n is node 1 to 9)"
 PRINT "(optional /X forces linux mode)"
 PRINT "(optional /Z ignores errors)"
 COLOR Plain, Black
 PRINT "Returning to System:"
 CALL Stop.Program
 END

' standard error trap for all Sic functions.
Error.Routine:
 PrinterLF = False
 Printing = False
 ErrorLine = Program.Line
 ErrorValue = ERR
 IF POS(0) > 1 THEN
    PRINT
 END IF
 Out2=Nul
 If Boot.Error=False Then
    CALL Display.Error
 Endif
 IF Program.Resume <= False THEN
    CALL Key.Prompt
 END IF
 DEF SEG
 InDEFSEG = False
 RESUME Error.Resume

' standard startup error trap.
Error.Routine2:
 If Program.Resume = False Then
    Print "SIC is terminating because of an error loading."
    Print "Returning to system with error:";
    Select Case Err
    Case 7
       Print " Out of memory."
       Print "Free more conventional Ram and restart."
    Case 14
       Print " Out of string space."
       Print "Free more conventional Ram and restart."
    Case Else
       Print Err
    End Select
 Endif
 If Program.Resume Then
    Print "Error";Err
 Endif
 Color 7, 0
 End

' check validity of program
SUB Analyze.Program(Var1, Var2, Var1$)
 Var2 = 0
 CALL Count.Lines(Last.Line)
 If Var1 = -1 Then
    COLOR Yellow, Black
    PRINT "Analyzing Program: ";Program.Name
    COLOR White, Black
    PRINT "Pass 1: counting matching structures."
 Endif
 RESTORE Analyze.Data1
 FOR Data.Count = 1 TO 8
    READ Start.Structure$, Stop.Structure$
    Nested.Count = False
    FOR Program.Line = 1 TO Last.Line
       Out2 = Program(Program.Line)
       Out2 = STRIM$(Out2)
       IF LEN(Out2) THEN
	  Temp1$ = STRIM$(Out2)
	  Temp1$ = UCASE$(Temp1$)
          Count.Data1 = False
	  IF LEFT$(Temp1$, LEN(Start.Structure$)) = Start.Structure$ THEN
             Count.Data1 = True
	  END IF
          IF Start.Structure$ = "FOR" THEN
             IF LEFT$(Temp1$, 5) = "FORIF" THEN
                Count.Data1 = False
             END IF
          END IF
          IF Count.Data1 THEN
	     Nested.Count = Nested.Count + 1
          END IF
	  IF Start.Structure$ = "IF" THEN
	     IF LEFT$(Temp1$, 6) = "END IF" THEN
		Nested.Count = Nested.Count - 1
		IF Nested.Count < False THEN
                   Var2 = -1
                   Var1$ = "Incomplete " + Start.Structure$
                   If Var1 = -1 Then
                      PRINT "Analyze error, pass 1: Incomplete "; Start.Structure$; ": Line:"; Program.Line
                   Endif
		   EXIT SUB
		END IF
	     END IF
	  END IF
          IF Stop.Structure$ = "LOOP" THEN
             IF LEFT$(Temp1$, 6) <> "LOOPIF" THEN
                Temp1$ = LEFT$(Temp1$, 4)
             END IF
          END IF
          IF Stop.Structure$ = "NEXT" THEN
             IF LEFT$(Temp1$, 6) <> "NEXTIF" THEN
                Temp1$ = LEFT$(Temp1$, 4)
             END IF
          END IF
          IF Stop.Structure$ = "END SELECT" THEN
             IF LEFT$(Temp1$, 12) <> "END SELECTIF" THEN
                Temp1$ = LEFT$(Temp1$, 10)
             END IF
          END IF
          IF Temp1$ = Stop.Structure$ THEN
	     Nested.Count = Nested.Count - 1
	     IF Nested.Count < False THEN
                Var2 = -1
                Var1$ = "Incomplete " + Start.Structure$
                If Var1 = -1 Then
                   PRINT "Analyze error, pass 1: Incomplete "; Start.Structure$; ": Line:"; Program.Line
                Endif
		EXIT SUB
	     END IF
	  END IF
       END IF
    NEXT
 NEXT
 If Var1 = -1 Then
    PRINT "Pass 2: counting matching imbedded structures."
 Endif
 RESTORE Analyze.Data2
 FOR Data.Count = 1 TO 3
    READ Start.Structure$, Stop.Structure$, Imbedded1$, Imbedded2$
    Nested.Count = False
    FOR Program.Line = 1 TO Last.Line
       Out2 = Program(Program.Line)
       Out2 = STRIM$(Out2)
       IF LEN(Out2) THEN
	  Temp1$ = STRIM$(Out2)
	  Temp1$ = UCASE$(Temp1$)
	  IF LEFT$(Temp1$, LEN(Start.Structure$)) = Start.Structure$ THEN
	     Nested.Count = Nested.Count + 1
	  END IF
	  IF LEFT$(Temp1$, LEN(Imbedded1$)) = Imbedded1$ THEN
	     Nested.Count = Nested.Count - 1
	     IF Nested.Count < False THEN
                Var2 = -1
                Var1$ = "Incomplete " + Start.Structure$
                If Var1 = -1 Then
                   PRINT "Analyze error, pass 2: Incomplete "; Start.Structure$; ": Line:"; Program.Line
                Endif
		EXIT SUB
	     END IF
	     Nested.Count = Nested.Count + 1
	  ELSE
             Temp2$ = Temp1$
             IF Imbedded2$ <> "ELSEIF" THEN
                IF LEFT$(Temp1$, LEN(Imbedded2$)) = Imbedded2$ THEN
                   Imbedded3 = INSTR(Temp1$, " ")
                   IF Imbedded3 = False THEN
                      Var2 = -1
                      Var1$ = "Bad " + Imbedded2$
                      If Var1 = -1 Then
                         PRINT "Analyze error, pass 2: Bad ";Imbedded2$; ": Line:";Program.Line
                      Endif
                      EXIT SUB
                   END IF
                   Temp2$ = LEFT$(Temp1$, Imbedded3 - 1)
                END IF
             END IF
             IF Temp2$ = Imbedded2$ THEN
		Nested.Count = Nested.Count - 1
		IF Nested.Count < False THEN
                   Var2 = -1
                   Var1$ = "Incomplete " + Start.Structure$
                   If Var1 = -1 Then
                      PRINT "Analyze error, pass 2: Incomplete "; Start.Structure$; ": Line:"; Program.Line
                   Endif
		   EXIT SUB
		END IF
		Nested.Count = Nested.Count + 1
	     END IF
	  END IF
	  IF Start.Structure$ = "IF" THEN
	     IF LEFT$(Temp1$, 6) = "END IF" THEN
		Nested.Count = Nested.Count - 1
		IF Nested.Count < False THEN
                   Var2 = -1
                   Var1$ = "Incomplete " + Start.Structure$
                   If Var1 = -1 Then
                      PRINT "Analyze error, pass 2: Incomplete "; Start.Structure$; ": Line:"; Program.Line
                   Endif
		   EXIT SUB
		END IF
	     END IF
	  END IF
          IF Temp1$ = Stop.Structure$ THEN
	     Nested.Count = Nested.Count - 1
	     IF Nested.Count < False THEN
                Var2 = -1
                Var1$ = "Incomplete " + Start.Structure$
                If Var1 = -1 Then
                   PRINT "Analyze error, pass 2: Incomplete "; Start.Structure$; ": Line:"; Program.Line
                Endif
		EXIT SUB
	     END IF
	  END IF
       END IF
    NEXT
 NEXT
 If Var1 = -1 Then
    PRINT "Pass 3: multipass structure match."
 Endif
 RESTORE Analyze.Data1
 FOR Data.Count = 1 TO 8
    READ Start.Structure$, Stop.Structure$
    FOR Program.Line = 1 TO Last.Line
       Out2 = Program(Program.Line)
       Out2 = STRIM$(Out2)
       IF LEN(Out2) THEN
	  Temp1$ = STRIM$(Out2)
	  Temp1$ = UCASE$(Temp1$)
          Count.Data1 = False
	  IF LEFT$(Temp1$, LEN(Start.Structure$)) = Start.Structure$ THEN
             Count.Data1 = True
          END IF
          IF Start.Structure$ = "FOR" THEN
             IF LEFT$(Temp1$, 5) = "FORIF" THEN
                Count.Data1 = False
             END IF
          END IF
          IF Count.Data1 THEN
             Nested.Count = 1
	     FOR Nested.Lines = Program.Line + 1 TO Last.Line
		Temp2$ = Program(Nested.Lines)
		Temp2$ = STRIM$(Temp2$)
		IF LEN(Temp2$) THEN
		   Temp2$ = STRIM$(Temp2$)
		   Temp2$ = UCASE$(Temp2$)
                   Count.Data2 = False
		   IF LEFT$(Temp2$, LEN(Start.Structure$)) = Start.Structure$ THEN
                      Count.Data2 = True
                   END IF
                   IF Start.Structure$ = "FOR" THEN
                      IF LEFT$(Temp2$, 5) = "FORIF" THEN
                         Count.Data2 = False
                      END IF
                   END IF
                   IF Count.Data2 THEN
		      Nested.Count = Nested.Count + 1
		   END IF
                   IF Stop.Structure$ = "LOOP" THEN
                      IF LEFT$(Temp2$, 6) <> "LOOPIF" THEN
                         Temp2$ = LEFT$(Temp2$, 4)
                      END IF
                   END IF
                   IF Stop.Structure$ = "NEXT" THEN
                      IF LEFT$(Temp2$, 6) <> "NEXTIF" THEN
                         Temp2$ = LEFT$(Temp2$, 4)
                      END IF
                   END IF
                   IF Stop.Structure$ = "END SELECT" THEN
                      IF LEFT$(Temp1$, 12) <> "END SELECTIF" THEN
                         Temp1$ = LEFT$(Temp1$, 10)
                      END IF
                   END IF
                   IF Temp2$ = Stop.Structure$ THEN
		      Nested.Count = Nested.Count - 1
                      IF Nested.Count = False THEN
                         EXIT FOR
                      END IF
		   END IF
		   IF Start.Structure$ = "IF" THEN
		      IF LEFT$(Temp2$, 6) = "END IF" THEN
			 Nested.Count = Nested.Count - 1
                         IF Nested.Count = False THEN
                            EXIT FOR
                         END IF
		      END IF
		   END IF
		END IF
	     NEXT
	     IF Nested.Count <> False THEN
                Var2 = -1
                Var1$ = "Mismatched " + Start.Structure$
                If Var1 = -1 Then
                   PRINT "Analyze error, pass 3: Mismatched "; Start.Structure$; ": Line:"; Program.Line
                Endif
		EXIT SUB
	     END IF
	  END IF
       END IF
    NEXT
 NEXT
 If Var1 = -1 Then
    PRINT "Pass 4: multipass imbedded struture match."
 Endif
 RESTORE Analyze.Data2
 FOR Data.Count = 1 TO 3
    READ Start.Structure$, Stop.Structure$, Imbedded1$, Imbedded2$
    FOR Program.Line = 1 TO Last.Line
       Out2 = Program(Program.Line)
       Out2 = STRIM$(Out2)
       IF LEN(Out2) THEN
	  Temp1$ = STRIM$(Out2)
	  Temp1$ = UCASE$(Temp1$)
	  IF LEFT$(Temp1$, LEN(Start.Structure$)) = Start.Structure$ THEN
	     Total.Nested1 = 1
	     Total.Nested2 = 0
	     Nested.Count1 = 1
	     Nested.Count2 = 0
	     FOR Nested.Lines = Program.Line + 1 TO Last.Line
		Temp2$ = Program(Nested.Lines)
		Temp2$ = STRIM$(Temp2$)
		IF LEN(Temp2$) THEN
		   Temp2$ = STRIM$(Temp2$)
		   Temp2$ = UCASE$(Temp2$)
		   IF LEFT$(Temp2$, LEN(Start.Structure$)) = Start.Structure$ THEN
		      Total.Nested1 = Total.Nested1 + 1
		      Nested.Count1 = Nested.Count1 + 1
		   END IF
                   IF Temp2$ = Stop.Structure$ THEN
		      Total.Nested2 = Total.Nested2 - 1
		      Nested.Count1 = Nested.Count1 - 1
                      IF Nested.Count1 = False THEN
			 EXIT FOR
		      END IF
		   END IF
		   IF Start.Structure$ = "IF" THEN
		      IF LEFT$(Temp2$, 6) = "END IF" THEN
			 Nested.Count1 = Nested.Count1 - 1
                         IF Nested.Count1 = False THEN
			    EXIT FOR
			 END IF
		      END IF
		   END IF
		   IF LEFT$(Temp2$, LEN(Imbedded1$)) = Imbedded1$ THEN
		      Nested.Count2 = Nested.Count2 + 1
		      IF Nested.Count2 > Nested.Count1 THEN
                         Var2 = -1
                         Var1$ = "Mismatched " + Start.Structure$
                         If Var1 = -1 Then
                            PRINT "Analyze error, pass 4: Mismatched "; Start.Structure$; ": Line:"; Program.Line
                         Endif
			 EXIT SUB
		      END IF
		      Nested.Count2 = Nested.Count2 - 1
		   ELSE
                      Temp3$ = Temp2$
                      IF Imbedded2$ <> "ELSEIF" THEN
                         IF LEFT$(Temp2$, LEN(Imbedded2$)) = Imbedded2$ THEN
                            Imbedded3 = INSTR(Temp2$, " ")
                            IF Imbedded3 = False THEN
                               Var2 = -1
                               Var1$ = "Bad " + Imbedded2$
                               If Var1 = -1 Then
                                  PRINT "Analyze error, pass 4: Bad ";Imbedded2$; ": Line:";Program.Line
                               Endif
                               EXIT SUB
                            END IF
                            Temp3$ = LEFT$(Temp2$, Imbedded3 - 1)
                         END IF
                      END IF
                      IF Temp3$ = Imbedded2$ THEN
			 Nested.Count2 = Nested.Count2 + 1
			 IF Nested.Count2 > Nested.Count1 THEN
                            Var2 = -1
                            Var1$ = "Mismatched " + Start.Structure$
                            If Var1 = -1 Then
                               PRINT "Analyze error, pass 4: Mismatched "; Start.Structure$; ": Line:"; Program.Line
                            Endif
			    EXIT SUB
			 END IF
			 Nested.Count2 = Nested.Count2 - 1
		      END IF
		   END IF
		END IF
	     NEXT
	     IF Total.Nested2 > Total.Nested1 THEN
                Var2 = -1
                Var1$ = "Mismatched " + Start.Structure$
                If Var1 = -1 Then
                   PRINT "Analyze error, pass 4: Mismatched "; Start.Structure$; ": Line:"; Program.Line
                Endif
		EXIT SUB
	     END IF
	  END IF
       END IF
    NEXT
 NEXT
 If Var1 = -1 Then
    PRINT "Pass 5: single pass imbedded control structure match."
 Endif
 RESTORE Analyze.Data3
 FOR Data.Count = 1 TO 5
    READ Start.Structure$, Stop.Structure$, Imbedded1$, Imbedded2$
    Nested.Count = False
    FOR Program.Line = 1 TO Last.Line
       Out2 = Program(Program.Line)
       Out2 = STRIM$(Out2)
       IF LEN(Out2) THEN
	  Temp1$ = STRIM$(Out2)
	  Temp1$ = UCASE$(Temp1$)
          Count.Data1 = False
	  IF LEFT$(Temp1$, LEN(Start.Structure$)) = Start.Structure$ THEN
             Count.Data1 = True
          END IF
          IF Start.Structure$ = "FOR" THEN
             IF LEFT$(Temp1$, 5) = "FORIF" THEN
                Count.Data1 = False
             END IF
          END IF
          IF Count.Data1 THEN
	     Nested.Count = Nested.Count + 1
	  END IF
          Count.Data2 = False
	  IF LEFT$(Temp1$, LEN(Stop.Structure$)) = Stop.Structure$ THEN
             Count.Data2 = True
          END IF
          IF Stop.Structure$ = "LOOP" THEN
             IF LEFT$(Temp1$, 6) = "LOOPIF" THEN
                Count.Data2 = False
             END IF
          END IF
          IF Stop.Structure$ = "NEXT" THEN
             IF LEFT$(Temp1$, 6) = "NEXTIF" THEN
                Count.Data2 = False
             END IF
          END IF
          IF Count.Data2 THEN
             Nested.Count = Nested.Count - 1
	  END IF
	  IF LEFT$(Temp1$, LEN(Imbedded1$)) = Imbedded1$ THEN
	     IF Nested.Count <= False THEN
                Var2 = -1
                Var1$ = "Badly nested " + Start.Structure$
                If Var1 = -1 Then
                   PRINT "Analyze error, pass 5: Badly nested "; Start.Structure$; ": Line:"; Program.Line
                Endif
		EXIT SUB
	     END IF
	  END IF
	  IF LEFT$(Temp1$, LEN(Imbedded2$)) = Imbedded2$ THEN
	     IF Nested.Count <= False THEN
                Var2 = -1
                Var1$ = "Badly nested " + Start.Structure$
                If Var1 = -1 Then
                   PRINT "Analyze error, pass 5: Badly nested "; Start.Structure$; ": Line:"; Program.Line
                Endif
		EXIT SUB
	     END IF
	  END IF
       END IF
    NEXT
 NEXT
 If Var1 = -1 Then
    COLOR Yellow, Black
    PRINT "Analysis completed. Program syntax correct."
 Endif
END SUB

' gets last line in program
SUB Count.Lines (Temp1%)
 Temp1% = False
 FOR Temp2% = Max.Lines TO 1 STEP -1
    Temp1$ = Program(Temp2%)
    Temp1$ = STRIM$(Temp1$)
    IF LEN(Temp1$) THEN
       EXIT FOR
    END IF
 NEXT
 Temp1% = Temp2%
END SUB

' counts lines in program
SUB Count.Lines2 (Temp1%)
 Temp1% = False
 FOR Temp2% = Max.Lines TO 1 STEP -1
    Temp1$ = Program(Temp2%)
    Temp1$ = STRIM$(Temp1$)
    IF LEN(Temp1$) THEN
       Temp1% = Temp1% + 1%
    END IF
 NEXT
END SUB

' reset file handles
SUB Decrease.PSP
 CLOSE
 InregsX.AX = &H6200
 CALL InterruptX(&H21, InregsX, OutregsX)
 PSP.Segment = OutregsX.BX
 DEF SEG = PSP.Segment
 POKE &H32, 20
 POKE &H34, &H18
 DEF SEG
 ' remove dimensioned arrays from memory
 ERASE Arrays, Strngs, Variables, GosubReturn, Definitions
 ERASE Program, Field.Array, File.Fields, Errors, LineBreak
 ERASE VariableBreak, VariableValue
 ERASE VariableBreak2, VariableValue2
 ERASE VariableBreak3, VariableValue3
END SUB

' displays error message
SUB Display.Error
 ' check error value range
 IF ErrorValue > 0 AND ErrorValue <= Max.Errors THEN
    Error.List$ = Errors(ErrorValue) ' store error string
 ELSE
    Error.List$ = "Error" + STR$(ErrorValue) ' store error number
 END IF
 ' parse out ': Line%1:'
 Parse = INSTR(Error.List$, "%1")
 IF Parse THEN
    ' check error line number
    IF ErrorLine = False THEN ' remove %1 string
       Error.List$ = LEFT$(Error.List$, Parse - 7) + MID$(Error.List$, Parse + 2)
    ELSE ' replace %1 string
       Error.List$ = LEFT$(Error.List$, Parse - 1) + STR$(ErrorLine) + MID$(Error.List$, Parse + 2)
    END IF
 END IF
 ' parse out ': %2.'
 Parse = INSTR(Error.List$, "%2")
 IF Parse THEN
    ' check token string
    IF Strng = CHR$(9) THEN
       Strng = "<tab>"
    END IF
    IF LEN(Strng) = False THEN ' remove %2 string
       Error.List$ = LEFT$(Error.List$, Parse - 1) + "<?>" + MID$(Error.List$, Parse + 2)
    ELSE ' replace %2 string
       Error.List$ = LEFT$(Error.List$, Parse - 1) + Strng + MID$(Error.List$, Parse + 2)
    END IF
 END IF
 ' check program loaded
 IF Program.Resume <= False THEN
    PRINT RTRIM$(Error.List$)
 ELSE
    ' display error for redirected in
    IF ControlBreak = False THEN
       PRINT "Error"; ErrorValue; " Line"; Program.Line
    END IF
 END IF
END SUB

' parses and stores a program line number
SUB Enter.Program.Line
 Assign = False
 Out2 = STRIM$(Out2)
 FOR Blanks = 1 TO LEN(Out2)
    Imbedded = False
    FOR Parse = 1 TO LEN(White.Space)
       IF MID$(Out2, Blanks, 1) = MID$(White.Space, Parse, 1) THEN
          Imbedded = Blanks
          EXIT FOR
       END IF
    NEXT
    IF Imbedded THEN
       Line.Number = INT(VAL(LEFT$(Out2, Imbedded - 1)))
       IF Line.Number >= 1 AND Line.Number <= Max.Lines THEN
          Out3 = MID$(Out2, Imbedded)
          IF STRIM$(Out3) = Nul THEN
             EXIT FOR
          END IF
          Assign = True
          Program(Line.Number) = Out3
          IF Program.Name = None THEN
             Program.Name = Untitled
	  END IF
       ELSE
          Out3 = STRIM$(Out2)
          IF INSTR("0123456789", LEFT$(Out3, 1)) THEN
             Assign = True
             PRINT "Line number out of range."
          END IF
       END IF
       EXIT SUB
    END IF
 NEXT
 Line.Number = INT(VAL(Out2))
 IF Line.Number > False AND Line.Number <= Max.Lines THEN
    Assign = True
    Program(Line.Number) = Nul
    CALL Count.Lines(Last.Line)
    IF Last.Line = False THEN
       Program.Name = None
    ELSE
       IF Program.Name = None THEN
          Program.Name = Untitled
       END IF
    END IF
 ELSE
    Out3 = STRIM$(Out2)
    IF INSTR("0123456789", LEFT$(Out3, 1)) THEN
       Assign = True
       PRINT "Line number out of range."
    END IF
 END IF
END SUB

' initiates a MS-DOS shell
SUB Shell.Program2(Var$)
 IF Var$ = Nul THEN
    COLOR 7, 0
    PRINT "Type 'EXIT' to return."
    CALL Execute.Command("")
    COLOR White, Black
    PRINT "Returned from DOS shell."
    EXIT SUB
 END IF

 X = FreeFileNumber
 FX$ = "SIC87A" + MID$(STR$(Node), 2) + ".BAT"
 Open FX$ For Output As #FileNumber
 Print #FileNumber, Var$
 Print #FileNumber, "PAUSE"
 Print #FileNumber, "EXIT"
 CLOSE #FileNumber
 FileNumber = False
 CALL Execute.Command(FX$)
 Close #FileNumber
 Kill FX$
 COLOR White, Black
 PRINT "Returned from DOS shell."
END SUB

' starts a MS-DOS shell
SUB Execute.Command (Shell.Statement$)
 DEF SEG ' restore basic segment
 Var = ClearBreak ' reset break flag

 ' store shell dta
 InregsX.AX = &H1A00
 InregsX.DS = VARSEG(ParseDTA)
 InregsX.DX = VARPTR(ParseDTA)
 CALL InterruptX(&H21, InregsX, OutregsX)

 ' store current drive/directory
 Current.Directory$ = CURDIR$

 ' start DOS command shell
 Shell.Command$ = ENVIRON$("COMSPEC")
 IF LEN(Shell.Statement$) THEN
    Shell.Command$ = Shell.Command$ + " /C " + Shell.Statement$
 END IF

 ' call shell routine
 SHELL Shell.Command$

 ' restore current drive/directory
 CHDRIVE Current.Directory$
 CHDIR Current.Directory$

 ' restore basic dta
 InregsX.AX = &H1A00
 InregsX.DS = BASIC.DTA.SEG
 InregsX.DX = BASIC.DTA.OFF
 CALL InterruptX(&H21, InregsX, OutregsX)

 Var = ClearBreak ' reset break flag
END SUB

' indents current program
SUB Indent.Program(Number.Spaces)
 CALL Analyze.Program(0, VarX, Var$)
 If VarX Then
    Color White
    Print "Program analyze error line:" + Str$(Program.Line) + ": " + Var$ + "."
    Exit Sub
 Endif
 Increment = INT(Number.Spaces)
 IF Number.Spaces = False THEN
    COLOR White, Black
    PRINT "Enter number of spaces to indent: ";
    Increment = INT(VAL(KeyboardLine$))
    PRINT
 END IF
 IF Increment <= 0 THEN
    EXIT SUB
 END IF
 Indent = False
 CALL Count.Lines(Last.Line)
 MX = LEN(STR$(Last.Line))
 FOR Program.Line = 1 TO Last.Line
    Number$ = SPACE$(MX - LEN(MID$(STR$(Program.Line), 2)))
    Out2 = Program(Program.Line)
    Out2 = STRIM$(Out2)
    IF LEN(Out2) THEN
       Temp1$ = STRIM$(Out2)
       Temp1$ = UCASE$(Temp1$)
       Temp1$ = TTRIM$(Temp1$, False)
       Next.Indent = False
       RESTORE Indent.Data
       DO
	  READ Keyword$, KeyIndent
	  IF Keyword$ = "EOF" THEN
	     EXIT DO
	  END IF
	  IF LEFT$(Temp1$, LEN(Keyword$)) = Keyword$ THEN
	     Next.Indent = KeyIndent
	     EXIT DO
	  END IF
       LOOP
       SELECT CASE Next.Indent
       CASE 1 ' right
          Out2 = SPACE$(Indent * Increment) + STRIM$(Out2)
	  Indent = Indent + 1
       CASE -1 ' left
	  Indent = Indent - 1
	  IF Indent < False THEN
	     PRINT "Indent error: Line:"; Program.Line
             EXIT SUB
	  END IF
          Out2 = SPACE$(Indent * Increment) + STRIM$(Out2)
       CASE -2 ' left/right
	  Indent = Indent - 1
	  IF Indent < False THEN
	     PRINT "Indent error: Line:"; Program.Line
             EXIT SUB
	  END IF
          Out2 = SPACE$(Indent * Increment) + STRIM$(Out2)
	  Indent = Indent + 1
       CASE ELSE
          Out2 = SPACE$(Indent * Increment) + STRIM$(Out2)
       END SELECT
       Out2 = Number$ + Out2
       Program(Program.Line) = Out2
    END IF
 NEXT
 IF Indent <> False THEN
    PRINT "Indent error: Line:"; Program.Line
    EXIT SUB
 END IF
 PRINT "Program indented."
END SUB

' deletes a .sic file
SUB Kill.Program(Temp.Name$)
 DO
    IF Temp.Name$ = Nul THEN
       COLOR White, Black
       PRINT "Program name to kill: ";
       Program.File$ = KeyboardLine$
       PRINT
    ELSE
       Program.File$ = Temp.Name$
    END IF
    IF LEN(Program.File$) = False THEN
       EXIT DO
    END IF
    Filename = Program.File$
    IF DIR$(Filename) = Nul THEN
       PRINT "File " + Quote + Filename + Quote + " does not exist."
       IF LEN(Temp.Name$) THEN
          EXIT DO
       END IF
    ELSE
       IF Temp.Name$ = Nul THEN
          COLOR White, Black
          PRINT "Kill " + Quote + Filename + Quote + ". "
          CALL More.Prompt("Are you sure(y/n)?", "yn", Output.Char$)
       ELSE
          Output.Char$ = "y"
       END IF
       COLOR White, Black
       IF Output.Char$ = "y" THEN
	  KILL Filename
          PRINT "Program " + Quote + Filename + Quote + " killed."
       ELSE
          PRINT "Program " + Quote + Filename + Quote + " not killed."
       END IF
       EXIT DO
    END IF
 LOOP
END SUB

' formats a double numeric string
FUNCTION FormatString$ (s#)
x$ = Nul
s$ = STR$(s#)
IF INSTR(s$, "D") THEN
   FormatString$ = s$
   EXIT FUNCTION
END IF
IF LEFT$(s$, 1) = "-" THEN
   e$ = "-"
   s$ = MID$(s$, 2)
END IF
s$ = LTRIM$(s$)
IF INSTR(s$, ".") THEN
   q$ = MID$(s$, INSTR(s$, "."))
   s$ = LEFT$(s$, INSTR(s$, ".") - 1)
END IF
FOR l = LEN(s$) TO 3 STEP -3
   x$ = MID$(s$, l - 2, 3) + "," + x$
NEXT
IF l > 0 THEN
   x$ = MID$(s$, 1, l) + "," + x$
END IF
IF LEN(s$) < 3 THEN
   x$ = s$
END IF
IF RIGHT$(x$, 1) = "," THEN
   x$ = LEFT$(x$, LEN(x$) - 1)
END IF
x$ = e$ + x$ + q$
FormatString$ = x$
END FUNCTION

' ambiguate function
FUNCTION Ambiguate$(Var$)
 DIM ASCIIZ1 AS STRING * 260
 DIM ASCIIZ2 AS STRING * 261
 DIM ParseDTA2 AS DTAtype

 ' store path
 IF RIGHT$(Var$, 1) = "\" THEN
    Var$ = Var$ + "*.*"
 END IF
 ASCIIZ1 = Var$ + CHR$(0)

 ' ambiguate filename
 InregsX.AX = &H7160
 InregsX.CX = &H8001
 InregsX.DS = VARSEG(ASCIIZ1)
 InregsX.SI = VARPTR(ASCIIZ1)
 InregsX.ES = VARSEG(ASCIIZ2)
 InregsX.DI = VARPTR(ASCIIZ2)
 CALL InterruptX(&H21, InregsX, OutregsX)

 ' check carry flag error
 Var2$ = Var$
 IF (OutregsX.Flags AND &H1) = &H0 THEN
    Var2$ = ASCIIZ2
    ' store windows 8.3 filename
    Imbedded = INSTR(Var2$, CHR$(0))
    IF Imbedded THEN
       Var2$ = LEFT$(Var2$, Imbedded - 1)
    END IF
 END IF

 ' return 8.3 filename
 Ambiguate$ = Var2$
END FUNCTION

' emulate shortfilename DIR$()
SUB FileDir (Var$)
 STATIC ParseDTA2 AS DTAtype

 ' restore directory search dta
 InregsX.AX = &H1A00
 InregsX.DS = VARSEG(ParseDTA2)
 InregsX.DX = VARPTR(ParseDTA2)
 CALL InterruptX(&H21, InregsX, OutregsX)

 IF LEN(Var$) THEN
    ' find first directory/filename
    ASCIIZ = Var$ + CHR$(0)
    InregsX.AX = &H4E00
    InregsX.CX = &H37 ' dir/file
    InregsX.DS = VARSEG(ASCIIZ)
    InregsX.DX = VARPTR(ASCIIZ)
    CALL InterruptX(&H21, InregsX, OutregsX)
 ELSE
    ' find next filename
    InregsX.AX = &H4F00
    CALL InterruptX(&H21, InregsX, OutregsX)
 END IF

 ' check findfirst error
 IF (OutregsX.Flags AND &H1) = &H1 THEN
    Out3 = Nul
 ELSE
    ' store filename
    Out3 = ParseDTA2.ASCIIZfilename
    IF INSTR(Out3, CHR$(0)) THEN
       Out3 = LEFT$(Out3, INSTR(Out3, CHR$(0)) - 1)
    END IF
 END IF

 ' restore basic dta
 InregsX.AX = &H1A00
 InregsX.DS = BASIC.DTA.SEG
 InregsX.DX = BASIC.DTA.OFF
 CALL InterruptX(&H21, InregsX, OutregsX)
END SUB

' emulate longfilename DIRX$()
SUB FileDirX(Var$)
 DIM ASCIIZ.File AS STRING * 260
 STATIC FileWDTA AS WDTAtype
 STATIC Wfile.Handle AS INTEGER

 ' init filename search
 IF LEN(Var$) THEN

    ' close long filename search
    IF WFile.Handle THEN
       InregsX.AX = &H71A1
       InregsX.BX = Wfile.Handle
       CALL InterruptX(&H21, InregsX, OutregsX)
    END IF

    ' store asciiz filename
    ASCIIZ.File = Var$ + CHR$(0)

    ' find first long filename
    InregsX.AX = &H714E
    InregsX.CX = &H37
    InregsX.SI = &H1
    InregsX.DS = VARSEG(ASCIIZ.File)
    InregsX.DX = VARPTR(ASCIIZ.File)
    InregsX.ES = VARSEG(FileWDTA)
    InregsX.DI = VARPTR(FileWDTA)
    CALL InterruptX(&H21, InregsX, OutregsX)
    Wfile.Handle = OutregsX.AX
 ELSE
    ' find next long filename
    InregsX.AX = &H714F
    InregsX.BX = Wfile.Handle
    InregsX.SI = &H1
    InregsX.ES = VARSEG(FileWDTA)
    InregsX.DI = VARPTR(FileWDTA)
    CALL InterruptX(&H21, InregsX, OutregsX)
 END IF

 ' check find first error
 IF (OutregsX.Flags AND &H1) = &H0 THEN
    ' store filename
    Out3 = FileWDTA.ASCIIZfull
    Out3 = LEFT$(Out3, INSTR(Out3, CHR$(0)) - 1)
    EXIT SUB
 END IF

 ' close long filename search
 Out3 = Nul
 IF WFile.Handle THEN
    InregsX.AX = &H71A1
    InregsX.BX = Wfile.Handle
    CALL InterruptX(&H21, InregsX, OutregsX)
    WFile.Handle = False
 END IF
END SUB

' lists specified .sic files
SUB List.Files (Var$, VarQ)
 ' VarQ = 0 list dirs only, VarQ = -1 list files and dirs

 ' declare file buffer
 DIM ParseDTA2 AS DTAtype
 DIM ASCIIZ2 AS STRING * 260, ASCIIZ3 AS STRING * 260

 ' declare file date and time work variables
 DIM File.Work.Date AS SINGLE, File.Work.Time AS SINGLE
 DIM HourTemp AS SINGLE, MinuteTemp AS SINGLE, SecondsTemp AS SINGLE
 DIM DayTemp AS SINGLE, MonthTemp AS SINGLE, YearTemp AS SINGLE
 DIM File.Date AS STRING * 10, File.Time AS STRING * 8

 ' declare file size work variables
 DIM File.Size AS DOUBLE, Total.Bytes AS DOUBLE
 DIM Total.Files AS DOUBLE, Total.Dirs AS DOUBLE

 ' reset counters
 Line.Count = 2
 Continuous.Display = False

 ' reset some variables
 Total.Bytes = Dfalse
 Total.Files = DFalse
 Total.Dirs = Dfalse

 ' store asciiz search filename
 V$ = CURDIR$
 IF RIGHT$(V$, 1) <> "\" THEN
    V$ = V$ + "\"
 END IF
 IF Var$ = Nul THEN
    Var$ = "*.sic"
    FileList$ = V$ + "*.SIC"
 ELSE
    IF INSTR(Var$, "\") = 0 THEN
       IF MID$(Var$, 2, 1) = ":" THEN
          IF LEN(Var$) = 2 THEN
             Var$ = Var$ + "\"
             FileList$ = Var$
          ELSE
             FileList$ = Var$
          END IF
       ELSE
          FileList$ = V$ + Var$
       END IF
    ELSE
       FileList$ = Var$
    END IF
    IF RIGHT$(Var$, 1) = "\" THEN
       Var$ = Var$ + "*.*"
       FileList$ = Var$
    END IF
 END IF
 ASCIIZ2 = Var$ + CHR$(0)

 ' display header
 COLOR White, Black
 PRINT "Searching: " + Quote + Var$ + Quote
 GOSUB Title.Header

 ' restore directory search dta
 InregsX.AX = &H1A00
 InregsX.DS = VARSEG(ParseDTA2)
 InregsX.DX = VARPTR(ParseDTA2)
 CALL InterruptX(&H21, InregsX, OutregsX)

 ' loop dirs/files
 FOR Var1 = 1 TO 2

    ' list dirs only
    IF Var1 = 2 THEN
       IF VarQ = 0 THEN
          EXIT FOR
       END IF
    END IF

    ' find first directory/filename
    InregsX.AX = &H4E00
    InregsX.CX = &H37 ' dir/file
    InregsX.DS = VARSEG(ASCIIZ2)
    InregsX.DX = VARPTR(ASCIIZ2)
    CALL InterruptX(&H21, InregsX, OutregsX)

    ' start file list loop
    DO
       ' check findfirst error
       IF (OutregsX.Flags AND &H1) = &H1 THEN
          EXIT DO
       END IF

       ' Bitfields for file attributes:

       ' Bits   Description
       '  7      shareable
       '  6      unused
       '  5      archive
       '  4      directory
       '
       '  3      volume label
       '  2      system
       '  1      hidden
       '  0      read-only

       ' store dir/file attribute
       Attribute% = ASC(ParseDTA2.FileAttr)
       Attr$ = Space$(6)
       IF (Attribute% AND &H20) = &H20 THEN
          Mid$(Attr$, 1, 1) = "A" ' archive
       END IF
       IF (Attribute% AND &H10) = &H10 THEN
          Mid$(Attr$, 2, 1) = "D" ' directory
       END IF
       IF (Attribute% AND &H4) = &H4 THEN
          Mid$(Attr$, 3, 1) = "S" ' system
       END IF
       IF (Attribute% AND &H2) = &H2 THEN
          Mid$(Attr$, 4, 1) = "H" ' hidden
       END IF
       IF (Attribute% AND &H1) = &H1 THEN
          Mid$(Attr$, 5, 1) = "R" ' read-only
       END IF

       ' determine search type
       VarX = 0
       IF (Attribute% AND &H10) = &H10 THEN
          IF Var1 = 1 THEN ' dir
             X$ = ParseDTA2.ASCIIZfilename
             IF INSTR(X$, CHR$(0)) THEN
                X$ = LEFT$(X$, INSTR(X$, CHR$(0)) - 1)
             END IF
             IF X$ = "." OR X$ = ".." THEN
                Eat$ = Nul
             ELSE
                ' count dirs
                Var4$ = FileList$
                For V = Len(Var4$) To 1 Step -1
                   If Mid$(Var4$, V, 1) = "\" Then
                      Var4$ = Mid$(Var4$, V + 1)
                      Exit For
                   Endif
                Next
                Var5$=Var4$
                Do
                   V=Instr(Var5$, ".")
                   If V Then
                      Var5$=Left$(Var5$,V-1)+Mid$(Var5$,V+1)
                   Else
                      Exit Do
                   Endif
                Loop
                ' string to search for = Var4$, Var5$
                ' string to search = X$
                '   not case-sensitive.
                CALL InstrSUB1(Flag1, Var4$, X$, -1)
                CALL InstrSUB1(Flag2, Var5$, X$, -1)
                IF Flag1 Or Flag2 THEN ' return match
                   FileSize$ = "<DIR>"
                   Total.Dirs = Total.Dirs + 1#
                   VarX = -1
                END IF
             END IF
          END IF
       ELSE
          IF Var1 = 2 THEN ' file
             VarX = -1
             ' count filesize
             Temp# = ASC(MID$(ParseDTA2.FileSize, 4, 1))
             Temp# = Temp# * &H100 + ASC(MID$(ParseDTA2.FileSize, 3, 1))
             Temp# = Temp# * &H100 + ASC(MID$(ParseDTA2.FileSize, 2, 1))
             Temp# = Temp# * &H100 + ASC(MID$(ParseDta2.FileSize, 1, 1))
             ' count bytes/files
             Total.Bytes = Total.Bytes + Temp#
             Total.Files = Total.Files + 1#
             CALL Suffix(Temp#, FileSize$) ' 1,024.0 KB
          END IF
       END IF

       ' select search type
       IF VarX THEN
          IF q THEN
             GOSUB Title.Header
             q = 0
          END IF

          ' store shortfilename
          Filename = ParseDTA2.ASCIIZfilename
          IF INSTR(Filename, CHR$(0)) THEN
             Filename = LEFT$(Filename, INSTR(Filename, CHR$(0)) - 1)
          END IF

          ' construct file date and time for display
          File.Work.Date = ASC(MID$(ParseDTA2.FileDate, 2, 1))
          File.Work.Date = File.Work.Date * &H100 + ASC(MID$(ParseDTA2.FileDate, 1, 1))

          YearTemp = INT(File.Work.Date / 512)
          MonthTemp = INT((File.Work.Date AND &H1E0) / 32)
          DayTemp = INT(File.Work.Date AND &H1F)

          File.Work.Time = ASC(MID$(ParseDTA2.FileTime, 2, 1))
          File.Work.Time = File.Work.Time * &H100 + ASC(MID$(ParseDTA2.FileTime, 1, 1))

          HourTemp = INT(File.Work.Time / 2048)
          MinuteTemp = INT((File.Work.Time AND &H7E0) / 32)
          SecondsTemp = INT((File.Work.Time AND &H1F) / 2)
          YearTemp = YearTemp + 1980

          File.Date = RIGHT$(STR$(MonthTemp + 100), 2) + "-" + RIGHT$(STR$(DayTemp + 100), 2) + "-" + MID$(STR$(YearTemp), 2)
          File.Time = RIGHT$(STR$(HourTemp + 100), 2) + ":" + RIGHT$(STR$(MinuteTemp + 100), 2) + ":" + RIGHT$(STR$(SecondsTemp + 100), 2)

          ' display shortfilename
          Display.Line$ = Filename + SPACE$(13 - LEN(Filename))
          COLOR Yellow, Black
          PRINT Display.Line$;

          ' display file date/time
          Display.Line$ = " " + File.Date + "  " + File.Time
          COLOR Green, Black
          PRINT Display.Line$;

          ' display file attribute
          COLOR Red, Black
          PRINT " "; Attr$;

          ' display file size
          COLOR Cyan, Black
          PRINT RTRIM$(FileSize$) ' 1,024.0 KB

          ' pause
          StartTime# = TIMER
          DO
             RunTime# = TIMER - StartTime#
             IF RunTime# < 0# THEN RunTime# = RunTime# + 86400#
             IF RunTime# >= .1 THEN EXIT DO
          LOOP

          ' check page length
          Line.Count = Line.Count + 1
          IF Line.Count >= PageLength THEN
             Line.Count = 2
             CALL More.Prompt(MorePrompt2, "yn" + CHR$(13), Output.Char$)
             SELECT CASE Output.Char$
             CASE "y", CHR$(13)
                q = -1
             CASE "n"
                EXIT FOR
             END SELECT
          END IF
       END IF

       ' restore directory search dta
       InregsX.AX = &H1A00
       InregsX.DS = VARSEG(ParseDTA2)
       InregsX.DX = VARPTR(ParseDTA2)
       CALL InterruptX(&H21, InregsX, OutregsX)

       ' find next filename
       InregsX.AX = &H4F00
       CALL InterruptX(&H21, InregsX, OutregsX)
    LOOP
 NEXT

 ' restore basic dta
 InregsX.AX = &H1A00
 InregsX.DS = BASIC.DTA.SEG
 InregsX.DX = BASIC.DTA.OFF
 CALL InterruptX(&H21, InregsX, OutregsX)

 ' check page length
 IF Line.Count >= PageLength - 3 THEN
    IF Continuous.Display = False THEN
       CALL More.Prompt(MorePrompt2, "yn" + CHR$(13), Output.Char$)
       IF Output.Char$ = "n" THEN
          EXIT SUB
       END IF
    END IF
 END IF

 ' check totals
 IF Total.Files = 0 AND Total.Dirs = 0 THEN
    COLOR Yellow, Black
    PRINT None
 END IF

 ' print totals
 COLOR White, Black
 PRINT "------------                             ----------"
 Total.Line$ = "Files " + FORMAT$(Total.Files, "#,##0")
 Total.Line$ = Total.Line$ + " Dirs " + FORMAT$(Total.Dirs, "#,##0")
 Total.Line$ = LEFT$(Total.Line$, 40)
 Total.Line$ = Total.Line$ + SPACE$(40 - LEN(Total.Line$))

 CALL Suffix(Total.Bytes, Var3$)
 If Len(Var3$) <= 11 Then
    Total.Line$ = Total.Line$ + SPACE$(11 - LEN(Var3$)) + Var3$
 ELSE
    Total.Line$ = Total.Line$ + Var3$
 END IF
 PRINT Total.Line$
 EXIT SUB

Title.Header:
 COLOR White, Black
 PRINT "Filename      Date        Time     Attr        Size"
 PRINT "------------  ----------  -------- ----- ----------"
 RETURN
END SUB

' lists specified drives
SUB List.Drives (Var$, VarQ)
 ' Var$ = "x..." only list drives in string
 ' VarQ = 0 list all drives, VarQ = 1 to 26 list drives in array list
 '    always skips A: and B: unless array element 1 or 2 is equal to 1.

 gosub drive.header
 l=0
 for c=1 to 26
    if Var$ <> Nul Then ' display specific drives
       x$=ucase$(Var$)
       If INSTR(x$, CHR$(c + 64)) then
          x=INSTR(x$, CHR$(c + 64))
          x=ASC(MID$(x$,x,1))
          IF x>=65 and x<=90 THEN
             x=x-64
             If c=x Then
                Gosub DisplayDrive
             Endif
          Endif
       Endif
    Else
       If VarQ = 0 Then ' display all drives except A: or B:
          If c>=3 Then
             Gosub DisplayDrive
          Endif
       Else
          If c<=2 Then ' check floppy override
             If Arrays(VarQ, c) = 1 Then
                Gosub DisplayDrive
             Endif
          Else
             If Arrays(VarQ, c) = 0 Then ' compare drive array list
                Gosub DisplayDrive
             Endif
          Endif
       Endif
    Endif
    if h=20 then
       h=0
       CALL More.Prompt(MorePrompt4, "y " + CHR$(13), Output.Char$)
       gosub drive.header
    endif   
 next
 if q=0 then
    Print "<none>"
 else
    Color White, Black
    Print "Total drives";l
 endif
 exit sub

DisplayDrive:
 Call Check.Drive(c, T#)
 if T# then
    h=h+1
    l=l+1
    q=-1
    c$=chr$(c+64)
    Color White, Black
    Print c$;":    ";
    Color Yellow, Black
    Call Vlabel(c)
    z$=Out3
    z$=z$+space$(13-len(z$))
    Print z$;
    Color Green, Black
    Call Vserial(c)
    Out3=Left$(Out3,9)
    If Out3="" Then Out3="<unknown>"
    Out3=Out3+Space$(11-Len(Out3))
    Print Out3;
    CALL Vtype(c)
    Out3 = LEFT$(Out3, 8)
    IF Out3 = "" THEN Out3 = "<n/a>"
    Out3 = Out3 + SPACE$(8 - LEN(Out3)) 
    PRINT Out3;
    Color Cyan, Black
    Out3=c$
    CALL TotalSpace(Out3)
    x# = INT(VAL(Out3))
    x1# = x#
    if x#>0# then
       CALL Suffix(x#, S$) ' 1,024.0 KB
       Print SPACE$(11 - LEN(S$)) + S$;
    else
       Print "      <n/a>";
    endif
    Out3=c$
    CALL FreeSpace(Out3)
    y# = INT(VAL(Out3))
    y1# = y#
    if y#>0# then
       CALL Suffix(y#, S$) ' 1,024.0 KB
       Print SPACE$(11 - LEN(S$)) + S$;
    else
       Print "      <n/a>";
    endif
    if x1#>0# or y1#>0# then
       z#=x1#-y1#
       CALL Suffix(z#, S$) ' 1,024.0 KB
       Print SPACE$(11 - LEN(S$)) + S$
    else
       Print "      <n/a>"
    endif
 endif
 return

drive.header:
 h=2
 Color White, Black
 PRINT "Drive Label        Serial     System        Total       Free       Used"
 PRINT "-----------------------------------------------------------------------"
 return
END SUB

' get drive freespace
SUB FreeSpace (Var$)
 ' check disk space.
 InregsX.AX = &H3600
 InregsX.DX = ASC(Ucase$(Var$)) - 64
 CALL InterruptX(&H21, InregsX, OutregsX)
 IF OutregsX.AX = &HFFFF THEN
    T# = 0#
 ELSE
'   free space on drive in bytes is AX * BX * CX
    T1# = OutregsX.AX : If T1# < 0# THEN T1# = T1# + 65536#
    T2# = OutregsX.BX : If T2# < 0# THEN T2# = T2# + 65536#
    T3# = OutregsX.CX : If T3# < 0# THEN T3# = T3# + 65536#
    T# = T1# * T2# * T3#
 END IF
 Var$ = Ltrim$(Str$(T#))
END SUB

' get drive totalspace
SUB TotalSpace (Var$)
 ' check disk space.
 InregsX.AX = &H3600
 InregsX.DX = ASC(Ucase$(Var$)) - 64
 CALL InterruptX(&H21, InregsX, OutregsX)
 IF OutregsX.AX = &HFFFF THEN
    T# = 0#
 ELSE
'   total space on drive in bytes is AX * CX * DX
    T1# = OutregsX.AX : If T1# < 0# THEN T1# = T1# + 65536#
    T2# = OutregsX.CX : If T2# < 0# THEN T2# = T2# + 65536#
    T3# = OutregsX.DX : If T3# < 0# THEN T3# = T3# + 65536#
    T# = T1# * T2# * T3#
 END IF
 Var$ = Ltrim$(Str$(T#))
END SUB

' display help menu
SUB List.Help
 DO
    COLOR White, Black
    PRINT "Help utility SIC v" + Version + " r" + Release + " list:"
    COLOR Yellow, Black
    PRINT "[1]command list"
    PRINT "[2]general documentation"
    PRINT "[3]formatting information"
    PRINT "[4]screen mode tables"
    PRINT "[5]error code values"
    PRINT "[6]syntax documentation"
    PRINT "[7]boolean syntax/charts"
    PRINT "[8]troolean syntax/charts"
    PRINT "[9]disclaimer notice"
    PRINT "[A]shareware information"
    CALL More.Prompt("Enter(1-A,Q to quit)?", "123456789aq", Output.Char$)
    SELECT CASE Output.Char$
    CASE "1"
       COLOR White, Black
       PRINT "SIC created: " + PublishDate
       COLOR Yellow, Black
       PRINT "ANALYZE  --  checks program structure."
       PRINT "CONTINUE --  continue halted program."
       PRINT "DEBUG    --  immediate debug mode."
       PRINT "FILES    --  display programs."
       PRINT "HELP     --  lists help topics."
       PRINT "INDENT   --  formats current program."
       PRINT "KILL     --  delete program."
       PRINT "LIST     --  display current program."
       PRINT "LOAD     --  read program from disk."
       PRINT "NEW      --  erase current program."
       PRINT "PRINT    --  print current program."
       PRINT "QUIT     --  exit SIC interpreter."
       PRINT "RENUMBER --  renumber program."
       PRINT "RUN      --  start current program."
       PRINT "SAVE     --  store current program."
       PRINT "SEARCH   --  search program for keyword."
       PRINT "SET      --  set string display for prompts."
       PRINT "SHELL    --  exit to dos shell."
       PRINT "WHATIS   --  enter immediate mode."
       CALL Key.Prompt
    CASE "2"
       Filename = "sic.doc"
       CALL List.Help.File
    CASE "3"
       Filename = "sicform.doc"
       CALL List.Help.File
    CASE "4"
       Filename = "sicscrn.doc"
       CALL List.Help.File
    CASE "5"
       Filename = "error.doc"
       CALL List.Help.File
    CASE "6"
       Filename = "syntax.doc"
       CALL List.Help.File
    CASE "7"
       Filename = "boolean.doc"
       CALL List.Help.File
    CASE "8"
       Filename = "troolean.doc"
       CALL List.Help.File
    CASE "9"
       Filename = "disclaim.doc"
       CALL List.Help.File
    CASE "a"
       Filename = "disclam2.doc"
       CALL List.Help.File
    CASE "q"
       EXIT DO
    END SELECT
 LOOP
END SUB

' displays a help file
SUB List.Help.File
 Call Find.File(Help.Filename$,File.Found)
 IF File.Found = False THEN
    PRINT "Helpfile " + Quote + Filename + Quote + " not found."
    PRINT "Set environment path variable to point at SIC files."
    CALL More.Prompt(MorePrompt, "ync", Output.Char$)
    EXIT SUB
 END IF
 X = FreeFileNumber
 OPEN Help.Filename$ FOR INPUT AS #FileNumber
 Continuous = False
 Line.Count = False
 DO WHILE NOT EOF(FileNumber)
    COLOR Yellow, Black
    LINE INPUT #FileNumber, Input.Line$
    PRINT Input.Line$
    IF Continuous = False THEN
       Line.Count = Line.Count + 1
       IF Line.Count >= PageLength THEN
	  Line.Count = False
          CALL More.Prompt(MorePrompt, "ync" + CHR$(13), Output.Char$)
	  SELECT CASE Output.Char$
          CASE "n"
	     EXIT DO
	  CASE "c"
	     Continuous = True
	  END SELECT
       END IF
    END IF
 LOOP
 CALL Key.Prompt
 CLOSE #FileNumber
 FileNumber = False
END SUB

' lists current .sic program
SUB List.Program(Start.Line, Stop.Line, Print.Prog)
 COLOR White, Black
 IF Print.Prog THEN
    CALL More.Prompt("Printer number(1-3,q)?", "123q", Output.Char$)
    IF Output.Char$ = "q" THEN
       EXIT SUB
    END IF
    Print.Number = INT(VAL(Output.Char$))
    X = FreeFileNumber
    OPEN "LPT"+MID$(STR$(Print.Number), 2) + ":" FOR OUTPUT AS #FileNumber
 END IF
 IF Print.Prog THEN
    PRINT #FileNumber, "Program: ";Program.Name
 ELSE
    PRINT "Program: ";Program.Name
 END IF
 Continuous.Display = False
 Line.Count = 1
 FOR Program.Line = Start.Line TO Stop.Line
    Out2 = Program(Program.Line)
    IF LEN(Out2) THEN
       COLOR Yellow, Black
       IF Print.Prog THEN
          IF INSTR(White.Space, LEFT$(Out2, 1)) THEN
             PRINT #FileNumber, MID$(STR$(Program.Line), 2) + Out2
          ELSE
             PRINT #FileNumber, MID$(STR$(Program.Line), 2) + " " + Out2
          END IF
       ELSE
          IF INSTR(White.Space, LEFT$(Out2, 1)) THEN
             V$ = ">" + MID$(STR$(Program.Line), 2) + Out2
          ELSE
             V$ = ">" + MID$(STR$(Program.Line), 2) + " " + Out2
          END IF
          PRINT V$
          IF Continuous.Display = False THEN
             Line.Count = Line.Count + INT((LEN(V$) - 1) / 80) + 1
             IF Line.Count >= PageLength THEN
                Line.Count = False
                CALL More.Prompt(MorePrompt, "ync" + CHR$(13), Output.Char$)
                SELECT CASE Output.Char$
                CASE "n"
                   EXIT FOR
                CASE "c"
                   Continuous.Display = True
                END SELECT
             END IF
          END IF
       END IF
    END IF
 NEXT
 COLOR White, Black
 PRINT "Program list ended."
 IF Print.Prog THEN
    CLOSE #FileNumber
    FileNumber = False
 END IF
END SUB

' searchs current .sic program
SUB Search.Program
 CALL Count.Lines(Last.Line)
 COLOR White, Black
 IF Last.Search.Line > False THEN
    CALL More.Prompt("Continue search(y/n)?", "yn", Output.Char$)
    IF Output.Char$ = "y" THEN
       Var1 = Last.Search.Case
       Var1$ = Last.Search.Keyword
       Start.Line = Last.Search.Line + 1
       IF Start.Line > Max.Lines THEN
          COLOR White, Black
          PRINT "Program search ended."
          EXIT SUB
       END IF
       PRINT "Continuing search from line:" + STR$(Start.Line)
       GOTO Start.Search
    END IF
 END IF
 DO
    Var = ClearBreak
    Program.Line = False
    Visible = 1
    COLOR Yellow, Black
    LOCATE , 1, 1
    PRINT "Search string?";
    LOCATE , , 1
    Out2 = RTRIM$(KeyboardLine$)
    PRINT
    IF BreakIS THEN
       COLOR Red, Black
       PRINT "*break*":SOUND 900,6
       SLEEP 1
    ELSE
       IF Out2 <> Nul THEN
          EXIT DO
       END IF
    END IF
 LOOP
 Var1$ = Out2
 CALL More.Prompt("Case-sensitive(y/n)?", "yn", Output.Char$)
 IF Output.Char$ = "y" THEN
    Var1 = True
 ELSE
    Var1 = False
 END IF
 Start.Line = 1

Start.Search:

 V$ = CURDIR$
 IF RIGHT$(V$, 1) <> "\" THEN
    V$ = V$ + "\"
 END IF
 PRINT "Searching Program: " + Quote + V$ + Program.Name + Quote

 FOR Program.Line = Start.Line TO Last.Line
    Out2 = Program(Program.Line)
    IF LEN(Out2) THEN
       CALL InstrSUB1(Flag, Var1$, Out2, Var1)
       IF Flag THEN
          COLOR Yellow, Black
          IF INSTR(White.Space, LEFT$(Out2, 1)) THEN
             PRINT ">" + MID$(STR$(Program.Line), 2) + Out2
          ELSE
             PRINT ">" + MID$(STR$(Program.Line), 2) + " " + Out2
          END IF
          CALL More.Prompt(MorePrompt2, "yn" + CHR$(13), Output.Char$)
          IF Output.Char$ = "n" THEN
             EXIT FOR
          END IF
       END IF
    END IF
 NEXT
 Last.Search.Case = Var1
 Last.Search.Line = Program.Line
 Last.Search.Keyword = Var1$
 COLOR White, Black
 PRINT "Program search ended."
END SUB

' loads a .sic file
SUB Load.Program(Temp.Name$)
 IF SaveOnExit THEN
    CALL Save.Current
 END IF
 IF Temp.Name$ = Nul THEN
    COLOR White, Black
    PRINT "Program name to load: ";
    Program.File$ = KeyboardLine$
    PRINT
 ELSE
    Program.File$ = Temp.Name$
 END IF
 IF Program.File$ = Nul THEN
    EXIT SUB
 END IF
 Call Concatenate(Program.File$)
 Filename = Program.File$
 Last.Search.Line = False
 CALL New.Program
 Program.Name = None
 CALL Read.Program
 Program.Name = Filename
 COLOR White, Black
 PRINT "Program loaded."
END SUB

' prompts for a keystroke
SUB Key.Prompt
 If Boot.Error = False Then
    COLOR White, Black
    LOCATE , , 1
    PRINT "Press any key to continue:";
    T# = TIMER
    DO
       ' display statusline
       T2# = TIMER - T#
       IF T2# < 0# THEN T2# = T2# + 86400#
       IF T2# >= 1# THEN
          T# = TIMER
          X = CSRLIN
          Y = POS(0)
          CALL StatusLine
          COLOR LastColor1, LastColor2
          COLOR Plain, Black
          LOCATE X, Y, 1
          PRINT "";
       END IF
       LOCATE , , 1
       IF KeyIs THEN
          Var$ = KeyboardChar$
          EXIT DO
       END IF
    LOOP
    PRINT
 Endif
END SUB

' prompts for multiple input characters
SUB More.Prompt (Input.String$, Input.Mask$, Output.String$)
 COLOR White, Black
 PRINT Input.String$; " ";
 Input.Char$ = Nul
 T# = TIMER
 DO
    ' display statusline
    T2# = TIMER - T#
    IF T2# < 0# THEN T2# = T2# + 86400#
    IF T2# >= 1# THEN
       T# = TIMER
       X = CSRLIN
       Y = POS(0)
       CALL StatusLine
       COLOR LastColor1, LastColor2
       COLOR Plain, Black
       LOCATE X, Y, 1
       PRINT "";
    END IF
    LOCATE , , 1
    IF KeyIs THEN
       Input.Char$ = LCASE$(KeyboardChar$)
       IF INSTR(Input.Mask$, Input.Char$) THEN
          IF Input.Char$ = CHR$(13) THEN
             PRINT
          ELSE
             PRINT Input.Char$
          END IF
          Output.String$ = Input.Char$
          EXIT DO
       END IF
    END IF
 LOOP
END SUB

' remove current .sic program from memory
SUB New.Program
 ' erase/redimension program code array
 ERASE Program
 REDIM Program(1 TO Max.Lines) AS STRING
END SUB

' end of .sic program routine
SUB End.Program
 If LastScreen1 <> 0 Then
    SCREEN 0, 1, 0, 0
 Else
    If LastScreen2 <> 1 Then
       SCREEN 0, 1, 0, 0
    Else
       If LastScreen3 <> 0 Or LastScreen4 <> 0 Then
          SCREEN 0, 1, 0, 0
       Endif
    Endif
 Endif
 If LastWidth1 <> 80 Then
    If LastWidth2 <> 25 Then
       WIDTH 80, 25
    Endif
 Endif
 COLOR Plain, Black
 LOCATE , , , 8, 8
 IF LineFeed THEN
    IF POS(0) > 1 THEN
       PRINT
    END IF
 END IF
 IF Program.Resume <= False THEN
    Var = ClearBuffer
 END IF
 DEF SEG
 InDEFSEG = False
END SUB

' quit the Sic program
SUB Quit.Program
 LineFeed = False
 IF SaveOnExit THEN
    CALL Save.Current
 END IF
 CALL Key.Prompt
 CALL End.Program
 PRINT "Exiting to system.."
 CALL Stop.Program
END SUB

' stop Sic program/exit to dos
SUB Stop.Program
 On Local Error Goto StopError
 ' remove temp files
 IF LEN(Prepare.Filename) THEN
    IF DIR$(Prepare.Filename)<>"" THEN
       KILL Prepare.Filename
    END IF
 END IF
 IF DIR$("SIC64.BAT")<>"" THEN
    KILL "SIC64.BAT"
 END IF
 IF DIR$(DRX + "FILE4.DIR")<>"" THEN
    KILL DRX + "FILE4.DIR"
 END IF
 IF DIR$(DRX + "FILE5.DIR")<>"" THEN
    KILL DRX + "FILE5.DIR"
 END IF

 ' restore basic dta
 InregsX.AX = &H1A00
 InregsX.DS = BASIC.DTA.SEG
 InregsX.DX = BASIC.DTA.OFF
 CALL InterruptX(&H21, InregsX, OutregsX)

 ' restore BASIC segment
 DEF SEG

 ' restore current drive/directory
 CHDRIVE CurrentDirectory2
 CHDIR CurrentDirectory2

 ' restore DOS PSP file handles
 CALL Decrease.PSP

 ' exit program
 CALL ClearStatusLine
StopExit:
 COLOR Plain, Black
 SYSTEM
StopError:
 Resume StopExit
END SUB

' loads a program from disk
SUB Read.Program
    Var = 0: Var2 = 0
    FileNumber = FreeFile
    If Dir$(Filename)<>"" Then
        Open Filename For Input As #FileNumber
        Do While Not EOF(FileNumber)
            Line Input #FileNumber, Out2
            Out2 = STRIM$(Out2)
            If Len(Out2) Then
                If Left$(Out2, 1) = "'" Or Left$(Out2, 1) = ";" Then
                    Eat$ = Nul
                Else
                    For Blanks = 1 To Len(Out2)
                        Imbedded$ = Mid$(Out2, Blanks, 1)
                        If Imbedded$ >= "0" And Imbedded$ <= "9" Then
                            Eat$ = Nul
                        Else
                            Line.Number = Int(Val(Left$(Out2, Blanks)))
                            If Line.Number > False And Line.Number <= Max.Lines Then
                                Program$(Line.Number) = Mid$(Out2, Blanks)
                                Var2 = -1
                            Else
                                Var = Var + 1
                            End If
                            Exit For
                        End If
                    Next
                End If
            End If
        Loop
    End If
    Close #FileNumber
    FileNumber = False
    If Var Then
        Print "Lines ignored:"; Var
    End If
    If Var2 = 0 Then
        Print "No program loaded."
    End If
END SUB

' gets standard input .sic file
SUB Read.Stdin
 On Local Error Goto StdinError
 ' check stdin length
 InregsX.AX = &H4202 ' eof
 InregsX.BX = 0 ' stdin
 InregsX.CX = 0
 InregsX.DX = 0
 Call InterruptX(&H21, InregsX, OutregsX)
 IF OutregsX.AX = 0 THEN ' no input found
    Var$ = Inkey$
    EXIT SUB
 END IF

 ' reset run type
 Program.Resume = 1
 CALL New.Program

 ' reset stdin start
 InregsX.AX = &H4200
 InregsX.BX = 0 ' stdin
 InregsX.CX = 0
 InregsX.DX = 0
 Call InterruptX(&H21, InregsX, OutregsX)

 ' loop while reading input
 DO

    ' read character from standard input
    InregsX.AX = &H3F00
    InregsX.BX = 0 ' stdin
    InregsX.CX = 1 ' char
    InregsX.DS = VARSEG(Pipe.Buffer)
    InregsX.DX = VARPTR(Pipe.Buffer)
    Call InterruptX(&H21, InregsX, OutregsX)
    If (OutregsX.Flags AND &H1) = &H1 Then
       Exit Do
    Endif
    If (OutregsX.Flags AND &H1) = &H0 Then
       If OutregsX.AX = 0 Then
          Exit Do
       Endif
       Char$ = Pipe.Buffer

       ' check linebreak character ('C' or DOS).
       IF (ASC(Char$)=10 AND LINUX) OR (ASC(Char$)=13 AND LINUX=0) THEN
          Out2 = STRIM$(Out2)
          FOR Blanks = 1 TO LEN(White.Space)
             Imbedded = INSTR(Out2, MID$(White.Space, Blanks, 1))
             IF Imbedded THEN
                Line.Number = INT(VAL(LEFT$(Out2, Imbedded - 1)))
                IF Line.Number > False AND Line.Number <= Max.Lines THEN
                   Out2 = MID$(Out2, Imbedded)
                   Out2 = STRIM$(Out2)
                   Program(Line.Number) = Out2
                   EXIT FOR
                END IF
             END IF
          NEXT
          Out2 = Nul
       ELSE
          IF (ASC(Char$)=10 AND LINUX=0) THEN
             Eat$=""
          ELSE
             Out2 = Out2 + Char$
          END IF
       END IF
    END IF
 LOOP

 ' run stdin program
 Run.Type=False
 StdinEnabled=-1
 CALL Run.Program(False)
 CALL ClearStatusLine
StdinResume:
 COLOR 7, 0
 SYSTEM
StdinError:
 Resume StdinResume
END SUB

' renumbers current .sic program
SUB Renumber.Program(Start.Value, Increment.Value)
 CALL Analyze.Program(0, VarX, Var$)
 If VarX Then
    Color White
    Print "Program analyze error line:" + Str$(Program.Line) + ": " + Var$ + "."
    Exit Sub
 Endif
 IF Start.Value = False OR Increment.Value = False THEN
    COLOR White, Black
    PRINT "Starting line number: ";
    Start.Line = INT(VAL(KeyboardLine$))
    PRINT
    PRINT "Increment value: ";
    Increment = INT(VAL(KeyboardLine$))
    PRINT
 ELSE
    Start.Line = Start.Value
    Increment = Increment.Value
 END IF
 Start.Line = INT(Start.Line)
 IF Start.Line <= False THEN
    PRINT "Bad start line value."
    EXIT SUB
 END IF
 IF Start.Line > Max.Lines THEN
    PRINT "Bad start line value."
    EXIT SUB
 END IF
 Increment = INT(Increment)
 IF Increment <= False THEN
    PRINT "Bad increment value."
    EXIT SUB
 END IF
 IF Increment > INT(Max.Lines / 2) THEN
    PRINT "Bad increment value."
    EXIT SUB
 END IF
 DO
    Temp.Filename$ = TempName$
    Filename = Temp.Filename$ + ".sc1"
    IF DIR$(Filename) = Nul THEN
       EXIT DO
    END IF
 LOOP
 CALL Store.Program
 CALL Count.Lines(Last.Line)
 REDIM Renumber.List(1 TO Max.Lines) AS INTEGER
 New.Line.Number = Start.Line
 FOR Line.Number = 1 TO Last.Line
    ProgramLine$ = Program(Line.Number)
    IF STRIM$(ProgramLine$) <> Nul THEN
       IF New.Line.Number > Max.Lines THEN
	  COLOR White, Black
	  PRINT "Renumber list exceeds"; STR$(Max.Lines); " lines."
	  ERASE Renumber.List
	  EXIT SUB
       END IF
       Renumber.List(New.Line.Number) = Line.Number
       New.Line.Number = New.Line.Number + Increment
    END IF
 NEXT
 PRINT "Renumbering program.."
 FOR Line.Number = 1 TO Last.Line
    New.Program.Line$ = Program(Line.Number)
    Old.Program.Line$ = Program(Line.Number)
    Old.Program.Line$ = STRIM$(Old.Program.Line$)
    Old.Program.Line$ = TTRIM$(Old.Program.Line$, True)
    IF Old.Program.Line$ <> Nul THEN
       Line.Renumbered = True
       IF UCASE$(LEFT$(Old.Program.Line$, 7)) = "RESTORE" THEN
          V$ = RTRIM$(MID$(Old.Program.Line$, 8))
          IF LEN(V$) THEN
             IF INT(VAL(V$)) > 0 THEN
                Line.Renumbered = False
                Imbedded = INSTR(UCASE$(New.Program.Line$), "RESTORE") + 7
                GOSUB Next.Space
                Number$ = MID$(Old.Program.Line$, 8)
                Number$ = STRIM$(Number$)
                Old.Line.Number = INT(VAL(Number$))
                IF Old.Line.Number > False AND Old.Line.Number <= Max.Lines THEN
                   FOR New.Line.Number = 1 TO Max.Lines
                      IF Renumber.List(New.Line.Number) = Old.Line.Number THEN
                         Program(Line.Number) = Next.Program.Line$ + LTRIM$(STR$(New.Line.Number))
                         Line.Renumbered = True
                         EXIT FOR
                      END IF
                   NEXT
                END IF
             END IF
          END IF
       END IF
       IF UCASE$(LEFT$(Old.Program.Line$, 4)) = "GOTO" THEN
	  Line.Renumbered = False
          Imbedded = INSTR(UCASE$(New.Program.Line$), "GOTO") + 4
          GOSUB Next.Space
	  Number$ = MID$(Old.Program.Line$, 5)
	  Number$ = STRIM$(Number$)
	  Old.Line.Number = INT(VAL(Number$))
          IF Old.Line.Number > False AND Old.Line.Number <= Max.Lines THEN
             FOR New.Line.Number = 1 TO Max.Lines
                IF Renumber.List(New.Line.Number) = Old.Line.Number THEN
                   Program(Line.Number) = Next.Program.Line$ + LTRIM$(STR$(New.Line.Number))
                   Line.Renumbered = True
                   EXIT FOR
                END IF
             NEXT
          END IF
       END IF
       IF UCASE$(LEFT$(Old.Program.Line$, 13)) = "ON ERROR GOTO" THEN
	  Line.Renumbered = False
          Imbedded = INSTR(UCASE$(New.Program.Line$), "GOTO") + 4
          GOSUB Next.Space
	  Number$ = MID$(Old.Program.Line$, 14)
	  Number$ = STRIM$(Number$)
	  Old.Line.Number = INT(VAL(Number$))
          IF Old.Line.Number > False AND Old.Line.Number <= Max.Lines THEN
             FOR New.Line.Number = 1 TO Max.Lines
                IF Renumber.List(New.Line.Number) = Old.Line.Number THEN
                   Program(Line.Number) = Next.Program.Line$ + LTRIM$(STR$(New.Line.Number))
                   Line.Renumbered = True
                   EXIT FOR
                END IF
             NEXT
          END IF
       END IF
       IF UCASE$(LEFT$(Old.Program.Line$, 6)) = "RESUME" THEN
	  IF UCASE$(LEFT$(Old.Program.Line$, 15)) <> "RESUME PREVIOUS" THEN
	     IF UCASE$(LEFT$(Old.Program.Line$, 11)) <> "RESUME SAME" THEN
		IF UCASE$(LEFT$(Old.Program.Line$, 11)) <> "RESUME NEXT" THEN
		   Line.Renumbered = False
                   Imbedded = INSTR(UCASE$(New.Program.Line$), "RESUME") + 6
                   GOSUB Next.Space
		   Number$ = MID$(Old.Program.Line$, 7)
		   Number$ = STRIM$(Number$)
		   Old.Line.Number = INT(VAL(Number$))
                   IF Old.Line.Number > False AND Old.Line.Number <= Max.Lines THEN
                      FOR New.Line.Number = 1 TO Max.Lines
                         IF Renumber.List(New.Line.Number) = Old.Line.Number THEN
                            Program(Line.Number) = Next.Program.Line$ + LTRIM$(STR$(New.Line.Number))
                            Line.Renumbered = True
                            EXIT FOR
                         END IF
                      NEXT
                   END IF
		END IF
	     END IF
	  END IF
       END IF
       IF UCASE$(LEFT$(Old.Program.Line$, 5)) = "GOSUB" THEN
	  Line.Renumbered = False
          Imbedded = INSTR(UCASE$(New.Program.Line$), "GOSUB") + 5
          GOSUB Next.Space
	  Number$ = MID$(Old.Program.Line$, 6)
	  Number$ = STRIM$(Number$)
	  Old.Line.Number = INT(VAL(Number$))
          IF Old.Line.Number > False AND Old.Line.Number <= Max.Lines THEN
             FOR New.Line.Number = 1 TO Max.Lines
                IF Renumber.List(New.Line.Number) = Old.Line.Number THEN
                   Program(Line.Number) = Next.Program.Line$ + LTRIM$(STR$(New.Line.Number))
                   Line.Renumbered = True
                   EXIT FOR
                END IF
             NEXT
          END IF
       END IF
       IF UCASE$(LEFT$(Old.Program.Line$, 2)) = "ON" THEN
	  IF UCASE$(LEFT$(Old.Program.Line$, 8)) <> "ON ERROR" THEN
	     Line.Renumbered = False
             IF INSTR(UCASE$(Old.Program.Line$), "GOTO") THEN
                Imbedded = INSTR(UCASE$(New.Program.Line$), "GOTO") + 4
                Old.Number1$ = LTRIM$(MID$(New.Program.Line$, Imbedded))
                GOSUB Next.Space
                New.Program.Line1$ = Next.Program.Line$
                Imbedded2 = INSTR(UCASE$(Old.Program.Line$), "GOTO")
                Old.Number2$ = MID$(Old.Program.Line$, Imbedded2 + 4)
		DO
                   Imbedded1 = INSTR(Old.Number1$, ",")
                   IF Imbedded1 THEN
                      New.Number1$ = LEFT$(Old.Number1$, Imbedded1 - 1)
                      Old.Number1$ = MID$(Old.Number1$, Imbedded1 + 1)
		   ELSE
                      New.Number1$ = Old.Number1$
                      Old.Number1$ = Nul
		   END IF
                   IF INSTR(White.Space, MID$(New.Number1$, 1, 1)) THEN
                      Imbedded = 1
                      New.Program.Line$ = New.Number1$
                      GOSUB Next.Space
                      New.Number1$ = Next.Program.Line$
                   ELSE
                      New.Number1$ = Nul
                   END IF
                   Imbedded2 = INSTR(Old.Number2$, ",")
                   IF Imbedded2 THEN
                      New.Number2$ = LEFT$(Old.Number2$, Imbedded2 - 1)
                      Old.Number2$ = MID$(Old.Number2$, Imbedded2 + 1)
		   ELSE
                      New.Number2$ = Old.Number2$
                      Old.Number2$ = Nul
		   END IF
		   Line.Renumbered = False
                   Old.Line.Number = INT(VAL(New.Number2$))
                   IF Old.Line.Number > False AND Old.Line.Number <= Max.Lines THEN
                      FOR New.Line.Number = 1 TO Max.Lines
                         IF Renumber.List(New.Line.Number) = Old.Line.Number THEN
                            New.Program.Line1$ = New.Program.Line1$ + New.Number1$ + LTRIM$(STR$(New.Line.Number)) + ","
                            Line.Renumbered = True
                            EXIT FOR
                         END IF
                      NEXT
                   END IF
                   IF Old.Number2$ = Nul THEN
		      EXIT DO
		   END IF
		   IF Line.Renumbered = False THEN
		      EXIT DO
		   END IF
		LOOP
                IF RIGHT$(New.Program.Line1$, 1) = "," THEN
                   New.Program.Line1$ = LEFT$(New.Program.Line1$, LEN(New.Program.Line1$) - 1)
                END IF
                Program(Line.Number) = New.Program.Line1$
	     ELSE
                IF INSTR(UCASE$(Old.Program.Line$), "GOSUB") THEN
                   Imbedded = INSTR(UCASE$(New.Program.Line$), "GOSUB") + 5
                   Old.Number1$ = LTRIM$(MID$(New.Program.Line$, Imbedded))
                   GOSUB Next.Space
                   New.Program.Line1$ = Next.Program.Line$
                   Imbedded2 = INSTR(UCASE$(Old.Program.Line$), "GOSUB")
                   Old.Number2$ = MID$(Old.Program.Line$, Imbedded2 + 5)
                   DO
                      Imbedded1 = INSTR(Old.Number1$, ",")
                      IF Imbedded1 THEN
                         New.Number1$ = LEFT$(Old.Number1$, Imbedded1 - 1)
                         Old.Number1$ = MID$(Old.Number1$, Imbedded1 + 1)
                      ELSE
                         New.Number1$ = Old.Number1$
                         Old.Number1$ = Nul
                      END IF
                      IF INSTR(White.Space, MID$(New.Number1$, 1, 1)) THEN
                         Imbedded = 1
                         New.Program.Line$ = New.Number1$
                         GOSUB Next.Space
                         New.Number1$ = Next.Program.Line$
                      ELSE
                         New.Number1$ = Nul
                      END IF
                      Imbedded2 = INSTR(Old.Number2$, ",")
                      IF Imbedded2 THEN
                         New.Number2$ = LEFT$(Old.Number2$, Imbedded2 - 1)
                         Old.Number2$ = MID$(Old.Number2$, Imbedded2 + 1)
                      ELSE
                         New.Number2$ = Old.Number2$
                         Old.Number2$ = Nul
                      END IF
                      Line.Renumbered = False
                      Old.Line.Number = INT(VAL(New.Number2$))
                      IF Old.Line.Number > False AND Old.Line.Number <= Max.Lines THEN
                         FOR New.Line.Number = 1 TO Max.Lines
                            IF Renumber.List(New.Line.Number) = Old.Line.Number THEN
                               New.Program.Line1$ = New.Program.Line1$ + New.Number1$ + LTRIM$(STR$(New.Line.Number)) + ","
                               Line.Renumbered = True
                               EXIT FOR
                            END IF
                         NEXT
                      END IF
                      IF Old.Number2$ = Nul THEN
                         EXIT DO
                      END IF
                      IF Line.Renumbered = False THEN
                         EXIT DO
                      END IF
                   LOOP
                   IF RIGHT$(New.Program.Line1$, 1) = "," THEN
                      New.Program.Line1$ = LEFT$(New.Program.Line1$, LEN(New.Program.Line1$) - 1)
                   END IF
                   Program(Line.Number) = New.Program.Line1$
                END IF
	     END IF
	  END IF
       END IF
       IF Line.Renumbered = False THEN
	  COLOR White, Black
          PRINT "Error renumbering program: line";Line.Number
          Filename = Temp.Filename$ + ".sc1"
	  CALL Read.Program
	  COLOR White, Black
	  PRINT "Program reloaded."
          KILL Filename
	  ERASE Renumber.List
	  EXIT SUB
       END IF
    END IF
 NEXT
 PRINT "Resequencing line numbers.."
 Filename = Temp.Filename$ + ".sc1"
 CALL Store.Program
 X = FreeFileNumber
 OPEN Filename FOR INPUT AS #FileNumber
 CALL New.Program
 WHILE NOT EOF(FileNumber)
    LINE INPUT #FileNumber, New.Program.Line$
    FOR Blanks = 1 TO LEN(White.Space)
       Imbedded = INSTR(New.Program.Line$, MID$(White.Space, Blanks, 1))
       IF Imbedded THEN
          EXIT FOR
       END IF
    NEXT
    Old.Line.Number = INT(VAL(LEFT$(New.Program.Line$, Imbedded - 1)))
    FOR New.Program.Line = 1 TO Max.Lines
       IF Renumber.List(New.Program.Line) = Old.Line.Number THEN
          Program(New.Program.Line) = MID$(New.Program.Line$, Imbedded)
	  EXIT FOR
       END IF
    NEXT
 WEND
 CLOSE #FileNumber
 FileNumber = False
 COLOR White, Black
 PRINT "Program renumbered."
 KILL Filename
 ERASE Renumber.List
 EXIT SUB

' locates string with following white spaces
Next.Space:
 Next.Program.Line$ = New.Program.Line$
 DO
    IF Imbedded >= LEN(Next.Program.Line$) THEN
       EXIT DO
    END IF
    IF INSTR(White.Space, MID$(Next.Program.Line$, Imbedded + 1, 1)) THEN
       Imbedded = Imbedded + 1
    ELSE
       EXIT DO
    END IF
 LOOP
 Next.Program.Line$ = LEFT$(Next.Program.Line$, Imbedded)
 RETURN
END SUB

' starts current .sic program
SUB Run.Program(Start.Line)
 IncludeRecurse = 0
 ' check to continue halted program.
 IF Run.Type THEN
    COLOR Plain, Black
    GOTO RunProg
 END IF

 ' close any files
 CLOSE

 ' try to release any memory
 Var! = FRE(-1)
 Var! = FRE("A")

 ' store current drive/directory
 CurrentDirectory = CURDIR$

 ' restore run variables
 Debug.Active = -1
 DataLine = 1
 DataNumber = False
 ErrorLine = False
 ErrorType = False
 ErrorValue = False
 Nested.Gosub = False
 Max.Gosubs = 10
 ScreenMode = 0
 ScreenHeight = 25
 ScreenWidth = 80
 Visible = 1
 COLOR Plain, Black

 ' clear runtime arrays
 REDIM GosubReturn(1 TO 10) AS INTEGER
 FOR Count1 = 1 TO 26
    Variables(Count1) = Dfalse
    Strngs(Count1) = Nul
    FOR Count2 = 1 TO Max.Arrays
       Arrays(Count1, Count2) = Dfalse
    NEXT
 NEXT
 FOR Count1 = 1 TO Max.Functions
    Definitions(Count1) = Nul
 NEXT

 ' start program processing loop.
RunProg:

 ' start program execution
 CALL Prepare.Program

 ' check starting line number
 CALL Count.Lines(Last.Line)

 ' check to continue halted program.
 IF Run.Type THEN
    ' restore last program line processed.
    Program.Line = Run.Line - 1
 ELSE
    ' store first program line.
    Program.Line = False
    IF Start.Line THEN
       IF Start.Line > False AND Start.Line <= Last.Line THEN
          IF LEN(STRIM$(Program(Start.Line))) THEN
             Program.Line = Start.Line - 1
          ELSE
             GOTO Run.Error
          END IF
       ELSE
          GOTO Run.Error
       END IF
    END IF
 END IF

 ' clear Control-Break flag
 Var = ClearBreak

 ' reset segment flag
 IF Run.Type THEN
    InDEFSEG = InDEFSEG2
 ELSE
    InDEFSEG = False
    InDEFSEG2 = False
 END IF

 ' reset colors
 IF Run.Type THEN
    COLOR LastColor1, LastColor2
 ELSE
    LastColor1 = Plain
    LastColor2 = Black
 END IF

 ' reset screen width
 IF Run.Type = -1 THEN
    WIDTH LastWidth1, LastWidth2
 ELSE
    LastWidth1 = 80
    LastWidth2 = 25
 END IF

 ' reset screen mode
 IF Run.Type = -1 THEN
    SCREEN LastScreen1, LastScreen2, LastScreen3, LastScreen4
 ELSE
    LastScreen1 = 0
    LastScreen2 = 1
    LastScreen3 = 0
    LastScreen4 = 0
 END IF

 ' reset cursor size
 IF Run.Type THEN
    LOCATE , , LastCursor1, LastCursor2, LastCursor3
 ELSE
    LastCursor1 = 1
    LastCursor2 = 8
    LastCursor3 = 8
 END IF

 ' start program flow
 Break.Flag1 = False
 Break.Flag2 = False
 Break.Flag3 = False
 Break.Flag4 = False
 Break.Flag5 = False
 Break.Flag6 = False
 Break.Flag7 = False

 ' check for break variables
 FOR VarBreak = 1 TO 26
    IF VariableBreak(VarBreak) THEN
       ' store first break variable
       Break.Flag3 = VarBreak
       EXIT FOR
    END IF
 NEXT

 ' check for break variables
 FOR VarBreak = 1 TO 26
    IF LEN(VariableBreak2(VarBreak))<>False THEN
       ' store first break variable
       Break.Flag5 = VarBreak
       EXIT FOR
    END IF
 NEXT

 ' check for break variables
 FOR VarBreak = 1 TO 26
    FOR VarBreak2 = 1 TO Max.Arrays
       IF VariableBreak3(VarBreak, VarBreak2) THEN
          ' store first break variable
          Break.Flag7 = VarBreak
          EXIT FOR
       END IF
    NEXT
    IF Break.Flag7 THEN
       EXIT FOR
    END IF
 NEXT

 ' start program loop
 T! = TIMER
 DO
    ' check Control-Break flag
    IF BreakIS THEN
       EXIT DO
    END IF

    ' display statusline
    IF TIMER - T! >= 1! THEN
       X = CSRLIN
       Y = POS(0)
       CALL StatusLine
       COLOR LastColor1, LastColor2
       COLOR Plain, Black
       LOCATE X, Y, 1
       PRINT "";
       T! = TIMER
    END IF

    ' increment program flow
    Program.Line = Program.Line + 1
    IF Program.Line > Last.Line THEN
       Program.Line = Max.Lines + 1
    END IF
    IF Program.Line > Max.Lines THEN
       EXIT DO
    END IF

    ' test break lines
    IF Debug.Active THEN
       IF LEN(Program(Program.Line))<>False THEN
          ' display debug list line
          IF ListLines THEN
             IF POS(0) > 1 THEN
                PRINT
             END IF
             Print "Debug line: " + Format$(Cdbl(Program.Line)) + "."
          END IF
          ' test debug break line
          IF LineBreak(Program.Line) THEN
             Break.Flag1 = True
             EXIT DO
          END IF
       END IF
    END IF

    ' process the program line
    ControlBreak = False
    Out2 = STRIM$(Program(Program.Line))
    IF LEN(Out2) THEN

       ' clear recurse flag
       Recurse = 0

       ' enter the parser
       CALL Enter.Equate

       ' test break variables
       IF Debug.Active THEN
          IF Break.Flag3 THEN
             GOSUB Test.Break1
             IF Break.Flag2 THEN
                EXIT DO
             END IF
          END IF
          IF Break.Flag5 THEN
             GOSUB Test.Break2
             IF Break.Flag4 THEN
                EXIT DO
             END IF
          END IF
          IF Break.Flag7 THEN
             GOSUB Test.Break3
             IF Break.Flag6 THEN
                EXIT DO
             END IF
          END IF
       END IF
    END IF

    ' check control-c
    IF ControlBreak THEN
       EXIT DO
    END IF
 LOOP

 ' store last program line processed.
 Run.Line = Program.Line

 ' store segment flag.
 InDEFSEG2 = InDEFSEG

 ' end program parsing
 CALL End.Program

 ' stop processing program
 Program.Line = Max.Lines

 ' display any break messages
 IF Break.Flag1 THEN
    COLOR Red, Black
    PRINT "*line break*"
    COLOR Yellow, Black
    PRINT "Line:" + STR$(Run.Line) + "."
 END IF
 IF Break.Flag2 THEN
    COLOR Red, Black
    PRINT "*variable break*"
    COLOR Yellow, Black
    PRINT "Line:" + STR$(Run.Line) + "."
    PRINT "Variable: " + CHR$(VarBreak + 64)
    PRINT "Test: " + Symbol$ + " " + LTRIM$(STR$(VariableValue(VarBreak)))
    PRINT "Value: " + LTRIM$(STR$(Variables(VarBreak)))
 END IF
 IF Break.Flag4 THEN
    COLOR Red, Black
    PRINT "*variable break*"
    COLOR Yellow, Black
    PRINT "Line:" + STR$(Run.Line) + "."
    PRINT "Variable: " + CHR$(VarBreak + 64) + "$"
    IF LEN(VariableValue2(VarBreak)) > 40 THEN
       PRINT "Test: " + Symbol$ + " " + Quote + LEFT$(VariableValue2(VarBreak), 40) + "..."
    ELSE
       PRINT "Test: " + Symbol$ + " " + Quote + VariableValue2(VarBreak) + Quote
    END IF
    IF LEN(Strngs(VarBreak)) > 40 THEN
       PRINT "Value: " + Quote + LEFT$(Strngs(VarBreak), 40) + "..."
    ELSE
       PRINT "Value: " + Quote + Strngs(VarBreak) + Quote
    END IF
 END IF
 IF Break.Flag6 THEN
    COLOR Red, Black
    PRINT "*variable break*"
    COLOR Yellow, Black
    PRINT "Line:" + STR$(Run.Line) + "."
    PRINT "Variable: " + CHR$(VarBreak + 64) + "(" + MID$(STR$(VarBreak2), 2) + ")"
    PRINT "Test: " + Symbol$ + " " + LTRIM$(STR$(VariableValue3(VarBreak, VarBreak2)))
    PRINT "Value: " + " " + LTRIM$(STR$(Arrays(VarBreak, VarBreak2)))
 END IF

 If ControlBreak Then
    Call DisplayBreak
 Endif

 ' restore current drive/directory/program
 CLOSE
 CHDRIVE CurrentDirectory
 CHDIR CurrentDirectory
 GOSUB Read.Prog
 EXIT SUB

' exit with error
Run.Error:
 CLOSE
 GOSUB Read.Prog
 ERROR 145
 EXIT SUB

' read original program
Read.Prog:
 CALL New.Program
 Filename = Prepare.Filename
 CALL Read.Program
 KILL Prepare.Filename
 RETURN

Test.Break1:
 ' start with first break variable
 FOR VarBreak = Break.Flag3 TO 26
    IF VariableBreak(VarBreak) THEN
       SELECT CASE VariableBreak(VarBreak)
       CASE -1
          IF Variables(VarBreak) = VariableValue(VarBreak) THEN
             Break.Flag2 = True
             Symbol$ = "="
             RETURN
          END IF
       CASE 1
          IF Variables(VarBreak) <> VariableValue(VarBreak) THEN
             Break.Flag2 = True
             Symbol$ = "<>"
             RETURN
          END IF
       CASE 2
          IF Variables(VarBreak) > VariableValue(VarBreak) THEN
             Break.Flag2 = True
             Symbol$ = ">"
             RETURN
          END IF
       CASE 3
          IF Variables(VarBreak) >= VariableValue(VarBreak) THEN
             Break.Flag2 = True
             Symbol$ = ">="
             RETURN
          END IF
       CASE 4
          IF Variables(VarBreak) < VariableValue(VarBreak) THEN
             Break.Flag2 = True
             Symbol$ = "<"
             RETURN
          END IF
       CASE 5
          IF Variables(VarBreak) <= VariableValue(VarBreak) THEN
             Break.Flag2 = True
             Symbol$ = "<="
             RETURN
          END IF
       END SELECT
    END IF
 NEXT
 RETURN

Test.Break2:
 ' start with first break variable
 FOR VarBreak = Break.Flag5 TO 26
    IF VariableBreak2(VarBreak) THEN
       SELECT CASE VariableBreak2(VarBreak)
       CASE -1
          IF Strngs(VarBreak) = VariableValue2(VarBreak) THEN
             Break.Flag4 = True
             Symbol$ = "="
             RETURN
          END IF
       CASE 1
          IF Strngs(VarBreak) <> VariableValue2(VarBreak) THEN
             Break.Flag4 = True
             Symbol$ = "<>"
             RETURN
          END IF
       CASE 2
          IF Strngs(VarBreak) > VariableValue2(VarBreak) THEN
             Break.Flag4 = True
             Symbol$ = ">"
             RETURN
          END IF
       CASE 3
          IF Strngs(VarBreak) >= VariableValue2(VarBreak) THEN
             Break.Flag4 = True
             Symbol$ = ">="
             RETURN
          END IF
       CASE 4
          IF Strngs(VarBreak) < VariableValue2(VarBreak) THEN
             Break.Flag4 = True
             Symbol$ = "<"
             RETURN
          END IF
       CASE 5
          IF Strngs(VarBreak) <= VariableValue2(VarBreak) THEN
             Break.Flag4 = True
             Symbol$ = "<="
             RETURN
          END IF
       END SELECT
    END IF
 NEXT
 RETURN

Test.Break3:
 ' start with first break variable
 FOR VarBreak = Break.Flag7 TO 26
    FOR VarBreak2 = 1 TO Max.Arrays
       IF VariableBreak3(VarBreak, VarBreak2) THEN
          SELECT CASE VariableBreak3(VarBreak, VarBreak2)
          CASE -1
             IF Arrays(VarBreak, VarBreak2) = VariableValue3(VarBreak, VarBreak2) THEN
                Break.Flag6 = True
                Symbol$ = "="
                RETURN
             END IF
          CASE 1
             IF Arrays(VarBreak, VarBreak2) <> VariableValue3(VarBreak, VarBreak2) THEN
                Break.Flag6 = True
                Symbol$ = "<>"
                RETURN
             END IF
          CASE 2
             IF Arrays(VarBreak, VarBreak2) > VariableValue3(VarBreak, VarBreak2) THEN
                Break.Flag6 = True
                Symbol$ = ">"
                RETURN
             END IF
          CASE 3
             IF Arrays(VarBreak, VarBreak2) >= VariableValue3(VarBreak, VarBreak2) THEN
                Break.Flag6 = True
                Symbol$ = ">="
                RETURN
             END IF
          CASE 4
             IF Arrays(VarBreak, VarBreak2) < VariableValue3(VarBreak, VarBreak2) THEN
                Break.Flag6 = True
                Symbol$ = "<"
                RETURN
             END IF
          CASE 5
             IF Arrays(VarBreak, VarBreak2) <= VariableValue3(VarBreak, VarBreak2) THEN
                Break.Flag6 = True
                Symbol$ = "<="
                RETURN
             END IF
          END SELECT
       END IF
    NEXT
 NEXT
 RETURN
END SUB

Sub DisplayBreak
 IF Program.Resume <= False THEN
    COLOR Red, Black
    Select Case ControlBreak
    Case -1
       PRINT "*break*"
    Case -2
       Print "*filebreak* (file" + Str$(File.Num) + ")"
    Case -3
       Print "*filerecordbreak* (file" + Str$(File.Num) + ") (record" + Str$(RecordNumber) + ")"
    End Select
    Sound 900,6:Sleep 1
 END IF
End Sub

' prompts to store current .sic program
SUB Save.Current
 IF Program.Name <> None THEN
    CALL More.Prompt("Save current program(y/n)?", "yn", Output.Char$)
    IF Output.Char$ = "y" THEN
       CALL Save.Program(Nul)
    END IF
 END IF
END SUB

' stores current .sic program
SUB Save.Program(Temp.Name$)
 DO
    IF Temp.Name$ <> Nul THEN
       Program.File$ = Temp.Name$
    END IF
    IF Temp.Name$ = Nul Then
       DO
          COLOR Yellow, Black
          PRINT "Program name to save";
          IF Program.Name <> Untitled THEN
             COLOR White, Black
             PRINT "("; Program.Name; ")";
          END IF
          PRINT "? ";
          Program.File$ = KeyboardLine$
          PRINT
          IF Program.File$ <> Nul THEN
             EXIT DO
          END IF
          IF Program.File$ = Nul THEN
             IF Program.Name <> Untitled THEN
                Program.File$ = Program.Name
                EXIT DO
             END IF
          END IF
       LOOP
    END IF
    Call Concatenate(Program.File$)
    Filename = Program.File$
    IF Temp.Name$ = Nul THEN
       COLOR Cyan, Black
       PRINT "Save program as " + Quote + Filename + Quote;
       CALL More.Prompt(MorePrompt3, "ynq", Output.Char$)
    ELSE
       Output.Char$ = "y"
    END IF
    SELECT CASE Output.Char$
    CASE "y"
       IF Temp.Name$ = Nul THEN
          Output.Char$ = "y"
          IF DIR$(Filename) <> Nul THEN
             COLOR Red, Black
             PRINT "Program already exists. ";
             CALL More.Prompt("Overwrite(y/n)?", "yn", Output.Char$)
          END IF
       ELSE
          Output.Char$ = "y"
       END IF
       IF Output.Char$ = "y" THEN
          ErrorValue = False
	  CALL Store.Program
          If ErrorValue = False Then
             Program.Name = Program.File$
             COLOR White, Black
             PRINT "Program " + Quote + Filename + Quote + " saved to disk."
          Endif
	  EXIT DO
       END IF
    CASE "q"
       COLOR White, Black
       PRINT "Program not saved to disk."
       EXIT DO
    END SELECT
 LOOP
END SUB

' writes out the current .sic program to file
SUB Store.Program
 X = FreeFileNumber
 OPEN Filename FOR OUTPUT AS #FileNumber
 CALL Count.Lines(Last.Line)
 FOR Line.Number = 1 TO Last.Line
    ProgramLine$ = Program(Line.Number)
    IF STRIM$(ProgramLine$) <> Nul THEN
       IF INSTR(White.Space, LEFT$(ProgramLine$, 1)) THEN
          PRINT #FileNumber, MID$(STR$(Line.Number), 2) + ProgramLine$
       ELSE
          PRINT #FileNumber, MID$(STR$(Line.Number), 2) + " " + ProgramLine$
       END IF
    END IF
 NEXT
 CLOSE #FileNumber
 FileNumber = False
END SUB

' immediate debugging prompt.
SUB Debug.Command
 ControlBreak = False
 COLOR White, Black
 PRINT "Type 'Quit' to exit Debug mode."
 PRINT "Or type ? for brief command list."
 DO
    Var = ClearBreak
    Visible = 1
    COLOR Yellow, Black
    LOCATE , 1, 1
    PRINT Prompt3;
    LOCATE , , 1
    Color 7,0
    Temp0 = 3
    Out2 = KeyboardLine2$
    PRINT
    IF BreakIS THEN
       COLOR Red, Black
       PRINT "*break*":SOUND 900,6
       SLEEP 1
       Out2 = Nul
    END IF

    ' parse command
    Var$ = UCASE$(Out2)

    ' exit debug
    IF Var$ = "QUIT" THEN
       EXIT DO
    END IF

    ' parse debug command
    IF Var$ = "LISTLINES" THEN
       Out2 = "LL"
    END IF
    IF LEFT$(Var$, 8) = "SETBREAK" THEN
       Out2 = "SB" + MID$(Var$, 9)
    END IF
    IF LEFT$(Var$, 10) = "CLEARBREAK" THEN
       Out2 = "CB" + MID$(Var$, 11)
    END IF
    IF Var$ = "LISTBREAKS" THEN
       Out2 = "LB"
    END IF
    IF Var$ = "CLEARALLBREAKS" THEN
       Out2 = "CA"
    END IF
    IF LEFT$(Var$, 11) = "SETVARIABLE" THEN
       Out2 = "SV" + MID$(Var$, 12)
    END IF
    IF LEFT$(Var$, 13) = "CLEARVARIABLE" THEN
       Out2 = "CV" + MID$(Var$, 14)
    END IF
    IF LEFT$(Var$, 14) = "TOGGLEVARIABLE" THEN
       Out2 = "TV" + MID$(Var$, 15)
    END IF
    IF Var$ = "LISTVARIABLES" THEN
       Out2 = "LV"
    END IF
    IF Var$ = "CLEARALLVARIABLES" THEN
       Out2 = "AV"
    END IF

    If Var$ = "LISTFILE" Then ' LF
       Out2 = "LF"
    End If
    If Var$ = "CLEARALLFILE" Then ' AF
       Out2 = "AF"
    End If
    If Left$(Var$, 10) = "CLEARFILE " Then ' CF
       Out2 = "CF" + Mid$(Var$, 11)
    End If
    If Left$(Var$, 8) = "SETFILE " Then ' SF
       Out2 = "SF" + Mid$(Var$, 9)
    End If

    If Var$ = "LISTRECORD" Then ' LR
       Out2 = "LR"
    End If
    If Var$ = "CLEARALLRECORD" Then ' AR
       Out2 = "AR"
    End If
    If Left$(Var$, 12) = "CLEARRECORD " Then ' CR
       Out2 = "CR" + Mid$(Var$, 13)
    End If
    If Left$(Var$, 10) = "SETRECORD " Then ' SR
       Out2 = "SR" + Mid$(Var$, 11)
    End If

    ' parse shortcuts
    SELECT CASE UCASE$(LEFT$(Out2, 2))
    Case "LR"
       GoSub ListRecordFile
    Case "CR"
       GoSub ClearRecord
    Case "AR"
       GoSub ClearAllRecord
    Case "SR"
       GoSub SetRecord

    Case "LF"
       GoSub ListBreakFile
    Case "CF"
       GoSub ClearFile
    Case "AF"
       GoSub ClearAllFile
    Case "SF"
       GoSub SetFile
    CASE "LL"
       IF LEN(Out2) = 2 THEN
          ListLines = NOT ListLines
          If ListLines Then
             Print "Debug: List lines on."
          Else
             Print "Debug: List lines off."
          Endif
       ELSE
          COLOR White, Black
          PRINT "Unknown command."
       END IF
    CASE "SB"
       GOSUB Set.Break
    CASE "CB"
       GOSUB Clear.Break
    CASE "LB"
       IF LEN(Out2) = 2 THEN
          GOSUB Display.Breaks
       ELSE
          COLOR White, Black
          PRINT "Unknown command."
       END IF
    CASE "CA"
       IF LEN(Out2) = 2 THEN
          GOSUB Clear.All.Breaks
       ELSE
          COLOR White, Black
          PRINT "Unknown command."
       END IF
    CASE "SV"
       GOSUB Set.Variable
    CASE "CV"
       GOSUB Clear.Variable
    CASE "TV"
       GOSUB Toggle.Variable
    CASE "LV"
       IF LEN(Out2) = 2 THEN
          GOSUB Display.Variables
       ELSE
          COLOR White, Black
          PRINT "Unknown command."
       END IF
    CASE "AV"
       IF LEN(Out2) = 2 THEN
          GOSUB Clear.All.Variables
       ELSE
          COLOR White, Black
          PRINT "Unknown command."
       END IF
    CASE ELSE
       ' list commands
       IF Out2 = "?" THEN
          COLOR Yellow, Black
          PRINT "Brief list of Debug commands:"
          PRINT "Command               Shortcut"
          PRINT "-------------------   ---------"
          COLOR White, Black
          PRINT "LISTLINES             LL"
          PRINT "SETBREAK <l>          SB <l>"
          PRINT "CLEARBREAK <l>        CB <l>"
          PRINT "LISTBREAKS            LB"
          PRINT "CLEARALLBREAKS        CA"
          PRINT "SETVARIABLE <x>,<n>   SV <x>,<n>"
          PRINT "CLEARVARIABLE <x>     CV <x>"
          PRINT "TOGGLEVARIABLE <x>    TV <x>"
          PRINT "LISTVARIABLES         LV"
          PRINT "CLEARALLVARIABLES     AV"
          ' new 04/06/2022
          Print "SETFILE               SF <x>"
          Print "CLEARFILE             CF <x>"
          Print "LISTFILE              LF"
          Print "CLEARALLFILE          AF"
          ' new 04/10/2023
          Print "SETRECORD             SR <x>"
          Print "CLEARRECORD           CR <x>"
          Print "LISTRECORD            LR"
          Print "CLEARALLRECORD        AR"
       ELSE
          ' parse whatis
          COLOR Plain, Black
          Out2 = TTRIM$(Out2, True)
          ' clear recurse flag
          Recurse = 0
          ' start the parser
          CALL Enter.Equate
          IF LineFeed THEN
             PRINT
          END IF
          If ControlBreak Then
             Call DisplayBreak
          Endif
       END IF
    END SELECT
 LOOP
 CALL End.Program
 COLOR White, Black
 PRINT "Debug mode ended."
 EXIT SUB

ClearRecord:
   Out2 = LTrim$(Mid$(Out2, 3))
   If Left$(Out2, 1) = "," Then Out2 = Mid$(Out2, 2)
   Z& = Int(Val(Out2))
   If Z& > 0& And Z& <= MaxLong Then
      For X = 1 To RecordCount
         If FileRecords(X) = Z& Then
            FileRecords(X) = False
            Print "File record" + Str$(X) + " cleared."
            Return
         End If
      Next
   End If
   Print "Unknown record."
   Return

ClearAllRecord:
   For VarZ = 1 To RecordCount
      FileRecords(VarZ) = 0&
   Next
   Print "All file records cleared."
   Return

SetRecord:
   Out2 = LTrim$(Mid$(Out2, 3))
   If Left$(Out2, 1) = "," Then Out2 = Mid$(Out2, 2)
   Z& = Int(Val(Out2))
   If Z& > 0& And Z& <= MaxLong Then
      RecordCount = RecordCount + 1
      FileRecords(RecordCount) = Z&
      Print "File record" + Str$(Z&) + " set."
      Return
   End If
   Print "Unknown record."
   Return

ListRecordFile:
   Var4$ = Nul
   VarX = 0
   For Z = 1 To RecordCount
      If FileRecords(Z) Then
         Var4$ = Var4$ + Str$(FileRecords(Z)) + ","
         VarX = -1
      End If
   Next
   If VarX Then
      Var4$ = Left$(Var4$, Len(Var4$) - 1) + "."
      Print "File number records:" + Var4$
   Else
      Print "No file records found."
   End If
   Return

ListBreakFile:
   Var4$ = Nul
   VarX = 0
   If FileTrap0 Then
      Var4$ = " 0,"
      VarX = -1
   End If
   For Z = 1 To MaxFiles
      If FileTraps(Z) Then
         Var4$ = Var4$ + Str$(Z) + ","
         VarX = -1
      End If
   Next
   If VarX Then
      Var4$ = Left$(Var4$, Len(Var4$) - 1) + "."
      Print "File number breaks:" + Var4$
   Else
      Print "No file breaks found."
   End If
   Return

ClearFile:
   Out2 = LTrim$(Mid$(Out2, 3))
   If Left$(Out2, 1) = "," Then Out2 = Mid$(Out2, 2)
   Z = Int(Val(Out2))
   If Z = 0 Then
      FileTrap0 = 0
      Print "File trap" + Str$(0) + " cleared."
      Return
   End If
   If Z > False And Z <= MaxFiles Then
      FileTraps(Z) = False
      Print "File trap" + Str$(Z) + " cleared."
      Return
   End If
   Print "Unknown file."
   Return

ClearAllFile:
   For VarZ = 1 To MaxFiles
      FileTraps(VarZ) = False
   Next
   FileTrap0 = 0
   Print "All file traps cleared."
   Return

SetFile:
   Out2 = LTrim$(Mid$(Out2, 3))
   If Left$(Out2, 1) = "," Then Out2 = Mid$(Out2, 2)
   Z = Int(Val(Out2))
   If Z = 0 Then
      FileTrap0 = -1
      Print "File trap" + Str$(0) + " set."
      Return
   End If
   If Z > False And Z <= MaxFiles Then
      FileTraps(Z) = True
      Print "File trap" + Str$(Z) + " set."
      Return
   End If
   Print "Unknown file."
   Return

Display.Variables:
 Continuous.Display = False
 Line.Count = False
 Line.Flag = False
 FOR Var = 1 TO 26
    ' display break variables 
    IF VariableBreak(Var) THEN
       ' check page length
       GOSUB Check.Title
       COLOR Yellow, Black
       PRINT " "; CHR$(Var + 64); "        ";
       BreakSign = VariableBreak(Var)
       GOSUB Get.Sign
       PRINT Symbol$; "   ";
       PRINT STR$(VariableValue(Var))
       Line.Flag = True
    END IF

    ' check page length
    IF Line.Count = PageLength - 2 THEN
       GOSUB Check.Page
       SELECT CASE Output.Char$
       CASE "n"
          RETURN
       CASE "c"
          Continuous.Display = True
       END SELECT
    END IF
 NEXT
 FOR Var = 1 TO 26
    ' display break variables 
    IF VariableBreak2(Var) THEN
       ' check page length
       GOSUB Check.Title
       COLOR Yellow, Black
       PRINT " "; CHR$(Var + 64); "$       ";
       BreakSign = VariableBreak2(Var)
       GOSUB Get.Sign
       PRINT Symbol$; "   ";
       IF LEN(VariableValue2(Var)) > 40 THEN
          PRINT Quote + LEFT$(VariableValue2(Var), 40) + "..."
       ELSE
          PRINT Quote + VariableValue2(Var) + Quote
       END IF
       Line.FLag = True
    END IF

    ' check page length
    IF Line.Count = PageLength - 2 THEN
       GOSUB Check.Page
       SELECT CASE Output.Char$
       CASE "n"
          RETURN
       CASE "c"
          Continuous.Display = True
       END SELECT
    END IF
 NEXT
 FOR Var = 1 TO 26
    FOR Var2 = 1 TO Max.Arrays
       ' display break variables 
       IF VariableBreak3(Var, Var2) THEN
          ' check page length
          GOSUB Check.Title
          COLOR Yellow, Black
          PRINT " "; CHR$(Var + 64); "("; MID$(STR$(Var2), 2); ")  ";
          IF Var2 > 99 THEN
             PRINT " ";
          ELSE
             IF Var2 > 9 THEN
                PRINT "  ";
             ELSE
                PRINT "   ";
             END IF
          END IF
          BreakSign = VariableBreak3(Var, Var2)
          GOSUB Get.Sign
          PRINT Symbol$; "   ";
          PRINT STR$(VariableValue3(Var, Var2))
          Line.Flag = True
       END IF

       ' check page length
       IF Line.Count = PageLength - 2 THEN
          GOSUB Check.Page
          SELECT CASE Output.Char$
          CASE "n"
             RETURN
          CASE "c"
             Continuous.Display = True
          END SELECT
       END IF
    NEXT
 NEXT
 IF Line.Flag = False THEN
    COLOR White, Black
    PRINT "No variable breaks set."
 END IF
 RETURN

Display.Breaks:
 Continuous.Display = False
 Line.Count = False
 Line.Flag = False
 FOR Var = 1 TO Max.Lines
    ' display break variables 
    IF LineBreak(Var) THEN
       ' check page length
       Line.Count = Line.Count + 1
       IF Continuous.Display = False THEN
          IF Line.Count = 1 THEN
             COLOR White, Black
             PRINT "Line Number"
             PRINT "-----------"
          END IF
       END IF
       COLOR Yellow, Black
       PRINT STR$(Var)
       Line.Flag = True
    END IF

    ' check page length
    IF Line.Count = PageLength - 2 THEN
       GOSUB Check.Page
       SELECT CASE Output.Char$
       CASE "n"
          RETURN
       CASE "c"
          Continuous.Display = True
       END SELECT
    END IF
 NEXT
 IF Line.Flag = False THEN
    COLOR White, Black
    PRINT "No line breaks set."
 END IF
 RETURN

Check.Page:
 Line.Count = False
 IF Continuous.Display = False THEN
    CALL More.Prompt(MorePrompt, "ync" + CHR$(13), Output.Char$)
 END IF
 RETURN

Check.Title:
 Line.Count = Line.Count + 1
 IF Continuous.Display = False THEN
    IF Line.Count = 1 THEN
       COLOR White, Black
       PRINT "Variable  Sign Value"
       PRINT "--------------------"
    END IF
 END IF
 RETURN

Get.Sign:
 SELECT CASE BreakSign
 CASE -1
    Symbol$ = "= "
 CASE 1
    Symbol$ = "<>"
 CASE 2
    Symbol$ = "> "
 CASE 3
    Symbol$ = ">="
 CASE 4
    Symbol$ = "< "
 CASE 5
    Symbol$ = "<="
 END SELECT
 RETURN

Set.Break:
 BreakLine = INT(VAL(MID$(Out2,3)))
 IF BreakLine >= False AND BreakLine <= Max.Lines THEN
    LineBreak(BreakLine) = True
    COLOR White, Black
    PRINT "Break line"+STR$(BreakLine)+" set."
    RETURN
 END IF
 COLOR White, Black
 PRINT "Unknown line."
 RETURN

Clear.Break:
 BreakLine = INT(VAL(MID$(Out2,3)))
 IF BreakLine >= False AND BreakLine <= Max.Lines THEN
    LineBreak(BreakLine) = False
    COLOR White, Black
    PRINT "Break line"+STR$(BreakLine)+" cleared."
    RETURN
 END IF
 COLOR White, Black
 PRINT "Unknown line."
 RETURN

Set.Variable:
 Var$ = MID$(Out2,3)
 Var$ = LTRIM$(Var$)
 Var2$ = LEFT$(Var$, 1)
 Var2$ = UCASE$(Var2$)
 IF Var2$>="A" AND Var2$<="Z" THEN
    BreakVar = ASC(Var2$)-64
    Var$ = MID$(Var$, 2)
    Var$ = LTRIM$(Var$)
    IF LEFT$(Var$, 1) = "(" THEN
       Imbedded = INSTR(Var$, ")")
       IF Imbedded THEN
          Var2$ = LEFT$(Var$, Imbedded - 1)
          Var2$ = MID$(Var2$, 2)
          BreakVar2 = INT(VAL(Var2$))
          IF BreakVar2 >= 1 AND BreakVar2 <= Max.Arrays THEN
             Var$ = MID$(Var$, Imbedded + 1)
             Var$ = LTRIM$(Var$)
             IF LEFT$(Var$, 1) = "," THEN
                BreakVar3# = VAL(MID$(Var$, 2))
                VariableBreak3(BreakVar, BreakVar2) = True
                VariableValue3(BreakVar, BreakVar2) = BreakVar3#
                COLOR White, Black
                PRINT "Break variable "+CHR$(BreakVar+64)+"("+MID$(STR$(BreakVar2),2)+")"+" set to "+LTRIM$(STR$(BreakVar3#))+"."
                RETURN
             END IF
          END IF
       END IF
    ELSE
       IF LEFT$(Var$, 1) = "$" THEN
          Var$ = MID$(Var$, 2)
          Var$ = LTRIM$(Var$)
          IF LEFT$(Var$, 1) = "," THEN
             BreakVar2$ = MID$(Var$, 2)
             VariableBreak2(BreakVar) = True
             VariableValue2(BreakVar) = BreakVar2$
             COLOR White, Black
             PRINT "Break variable "+CHR$(BreakVar+64)+"$ set to "+Quote+BreakVar2$+Quote+"."
             RETURN
          END IF
       ELSE
          IF LEFT$(Var$, 1) = "," THEN
             BreakVar2# = VAL(MID$(Var$, 2))
             VariableBreak(BreakVar) = True
             VariableValue(BreakVar) = BreakVar2#
             COLOR White, Black
             PRINT "Break variable "+CHR$(BreakVar+64)+" set to "+LTRIM$(STR$(BreakVar2#))+"."
             RETURN
          END IF
       END IF
    END IF
 END IF
 COLOR White, Black
 PRINT "Unknown variable."
 RETURN

Clear.Variable:
 Var$ = MID$(Out2,3)
 Var$ = UCASE$(Var$)
 Var$ = RTRIM$(Var$)
 Var$ = LTRIM$(Var$)
 Var2$ = LEFT$(Var$, 1)
 IF Var2$>="A" AND Var2$<="Z" THEN
    BreakVar=ASC(Var2$)-64
    IF LEN(Var$) = 1 THEN
       VariableBreak(BreakVar) = False
       VariableValue(BreakVar) = DFalse
       COLOR White, Black
       PRINT "Break variable "+CHR$(BreakVar+64)+" cleared."
       RETURN
    ELSE
       IF LEN(Var$) = 2 THEN
          Var2$ = MID$(Var$, 2)
          IF Var2$ = "$" THEN
             VariableBreak2(BreakVar) = False
             VariableValue2(BreakVar) = Nul
             COLOR White, Black
             PRINT "Break variable "+CHR$(BreakVar+64)+"$ cleared."
             RETURN
          END IF
       ELSE
          Var$ = MID$(Var$, 2)
          IF LEFT$(Var$, 1) = "(" THEN
             Imbedded = INSTR(Var$, ")")
             IF Imbedded THEN
                Var$ = LEFT$(Var$, Imbedded - 1)
                Var$ = MID$(Var$, 2)
                BreakVar2 = VAL(Var$)
                IF BreakVar2>=1 AND BreakVar2<=Max.Arrays THEN
                   VariableBreak3(BreakVar, BreakVar2) = False
                   VariableValue3(BreakVar, BreakVar2) = DFalse
                   COLOR White, Black
                   PRINT "Break variable "+CHR$(BreakVar+64)+"("+MID$(STR$(BreakVar2),2)+") cleared."
                   RETURN
                END IF
             END IF
          END IF
       END IF
    END IF
 END IF
 COLOR White, Black
 PRINT "Unknown variable."
 RETURN

Toggle.Variable:
 Var$ = MID$(Out2,3)
 Var$ = UCASE$(Var$)
 Var$ = RTRIM$(Var$)
 Var$ = LTRIM$(Var$)
 Var2$ = LEFT$(Var$, 1)
 IF Var2$>="A" AND Var2$<="Z" THEN
    BreakVar=ASC(Var2$)-64
    IF LEN(Var$) = 1 THEN
       IF VariableBreak(BreakVar) THEN
          NewBreak = VariableBreak(BreakVar)
          GOSUB Next.Break
          VariableBreak(BreakVar) = NewBreak
          COLOR White, Black
          PRINT "Break variable "+CHR$(BreakVar+64)+" toggled to "; Symbol$; "."
          RETURN
       END IF
    ELSE
       IF LEN(Var$) = 2 THEN
          Var2$ = MID$(Var$, 2)
          IF Var2$ = "$" THEN
             IF VariableBreak2(BreakVar) THEN
                NewBreak = VariableBreak2(BreakVar)
                GOSUB Next.Break
                VariableBreak2(BreakVar) = NewBreak
                COLOR White, Black
                PRINT "Break variable "+CHR$(BreakVar+64)+"$ toggled to "; Symbol$; "."
                RETURN
             END IF
          END IF
       ELSE
          Var$ = MID$(Var$, 2)
          IF LEFT$(Var$, 1) = "(" THEN
             Imbedded = INSTR(Var$, ")")
             IF Imbedded THEN
                Var$ = LEFT$(Var$, Imbedded - 1)
                Var$ = MID$(Var$, 2)
                BreakVar2 = VAL(Var$)
                IF BreakVar2>=1 AND BreakVar2<=Max.Arrays THEN
                   IF VariableBreak3(BreakVar, BreakVar2) THEN
                      NewBreak = VariableBreak3(BreakVar, BreakVar2)
                      GOSUB Next.Break
                      VariableBreak3(BreakVar, BreakVar2) = NewBreak
                      COLOR White, Black
                      PRINT "Break variable "+CHR$(BreakVar+64)+"("+MID$(STR$(BreakVar2),2)+") toggled to "; Symbol$; "."
                      RETURN
                   END IF
                END IF
             END IF
          END IF
       END IF
    END IF
 END IF
 COLOR White, Black
 PRINT "Unknown variable."
 RETURN

Next.Break:
 SELECT CASE NewBreak
 CASE -1 ' =
    NewBreak = 1
    Symbol$ = "<>"
 CASE 1  ' <>
    NewBreak = 2
    Symbol$ = ">"
 CASE 2  ' >
    NewBreak = 3
    Symbol$ = ">="
 CASE 3  ' >=
    NewBreak = 4
    Symbol$ = "<"
 CASE 4  ' <
    NewBreak = 5
    Symbol$ = "<="
 CASE 5  ' <=
    NewBreak = True
    Symbol$ = "="
 END SELECT
 RETURN

Clear.All.Breaks:
 FOR BreakLine = 1 TO Max.Lines
    LineBreak(BreakLine) = False
 NEXT
 COLOR White, Black
 PRINT "All line breaks cleared."
 RETURN

Clear.All.Variables:
 FOR BreakVar = 1 TO 26
    VariableBreak(BreakVar) = False
    VariableValue(BreakVar) = DFalse
    VariableBreak2(BreakVar) = False
    VariableValue2(BreakVar) = Nul
    FOR BreakVar2 = 1 TO Max.Arrays
       VariableBreak3(BreakVar, BreakVar2) = False
       VariableValue3(BreakVar, BreakVar2) = DFalse
    NEXT
 NEXT
 COLOR White, Black
 PRINT "All variable breaks cleared."
 RETURN
END SUB

' immediate parsing prompt.
SUB Whatis.Command(What.Value$)
 ErrorType = False
 ErrorValue = False
 IF LEN(What.Value$) THEN
    Program.Line = False
    Out2 = TTRIM$(What.Value$, True)
    ' clear recurse flag
    Recurse = 0
    ' start the parser
    CALL Enter.Equate
    IF LineFeed THEN
       PRINT
    END IF
 ELSE
    COLOR White, Black
    PRINT "Type 'Quit' to exit Whatis mode."
    DO
       Var = ClearBreak
       Program.Line = False
       Visible = 1
       COLOR Yellow, Black
       LOCATE , 1, 1
       PRINT Prompt2;
       LOCATE , , 1
       Color 7,0
       Temp0 = 2
       Out2 = RTRIM$(KeyboardLine2$)
       PRINT
       IF BreakIS THEN
          COLOR Red, Black
          PRINT "*break*"
          SOUND 900,6
          SLEEP 1
       ELSE
          IF UCASE$(Out2) = "QUIT" THEN
             EXIT DO
          END IF
          IF Out2 <> Nul THEN
             COLOR Plain, Black
             Out2 = TTRIM$(Out2, True)
             CALL Trim.Comment

             ' clear recurse flag
             Recurse = 0

             ' start the parser
             CALL Enter.Equate
             IF LineFeed THEN
                PRINT
             END IF
             If ControlBreak Then
                Call DisplayBreak
             Endif
          END IF
       END IF
    LOOP
 END IF
 CALL End.Program
 COLOR White, Black
 PRINT "Whatis mode ended."
END SUB

' prepares program for analyze/run command
SUB Prepare.Program
 ' write out current program
 DO
    Temp.Filename$ = TempName$
    Filename = DRX$ + Temp.Filename$ + ".sc2"
    IF DIR$(Filename) = Nul THEN
       EXIT DO
    END IF
 LOOP
 Prepare.Filename = Filename
 CALL Store.Program

 ' remove remark at end of line,
 '   search for end of matching double quotes
 '   which might contain an apostrophe.
 CALL Count.Lines(Last.Line)
 FOR Program.Line = 1 TO Last.Line
    Out2 = STRIM$(Program(Program.Line))
    IF LEN(Out2) THEN
       Start.Char = 1
       DO
          Quote.Start = False
          FOR Temp = Start.Char TO LEN(Out2)
             IF MID$(Out2, Temp, 1) = CHR$(34) THEN
                Quote.Start = Temp
                EXIT FOR
             END IF
          NEXT
          IF Quote.Start = False THEN
             EXIT DO
          END IF
          Quote.Stop = False
          FOR Temp2 = Quote.Start + 1 TO LEN(Out2)
             IF MID$(Out2, Temp2, 1) = CHR$(34) THEN
                Quote.Stop = Temp2
                EXIT FOR
             END IF
          NEXT
          IF Quote.Stop = False THEN
             EXIT DO
          END IF
          Start.Char = Quote.Stop + 1
       LOOP
       FOR Temp = Start.Char TO LEN(Out2)
          IF MID$(Out2, Temp, 1) = "'" THEN
             Out2 = LEFT$(Out2, Temp - 1)
             EXIT FOR
          END IF
       NEXT
       Out2 = STRIM$(Out2)
       IF LEN(Out2) THEN
          Program(Program.Line) = Out2
       ELSE
          Program(Program.Line) = "REM"
       END IF
    END IF
 NEXT

 ' concatenate line continuation statements,
 '   remove continued lines.
 Program.Line = False
 CALL Count.Lines(Last.Line)
 DO
    Program.Line = Program.Line + 1
    IF Program.Line > Last.Line THEN
       EXIT DO
    END IF
    IF Program.Line > Max.Lines THEN
       EXIT DO
    END IF
    First.Line = Program.Line
    Out2 = STRIM$(Program(Program.Line))
    IF LEN(Out2) THEN
       DO
          IF RIGHT$(Out2, 1) = "_" THEN
             DO
                V$ = RIGHT$(Out2, 1)
                IF V$ = CHR$(9) OR V$ = CHR$(32) OR V$ = "_" THEN
                   Out2 = LEFT$(Out2, LEN(Out2) - 1)
                ELSE
                   EXIT DO
                END IF
             LOOP
             FOR Next.Line = Program.Line + 1 TO Last.Line
                Out3 = STRIM$(Program(Next.Line))
                IF LEN(Out3) THEN
                   Out2 = Out2 + Out3
                   Program.Line = Next.Line
                   Program(Next.Line) = "REM"
                   EXIT FOR
                END IF
             NEXT
          ELSE
             EXIT DO
          END IF
          Out2 = STRIM$(Out2)
       LOOP
       Program(First.Line) = Out2
    END IF
 LOOP

 ' compress program array,
 '   replaces all white spaces with single space,
 '   replaces all double spaces with single space.
 Program.Line = False
 CALL Count.Lines(Last.Line)
 FOR Program.Line = 1 TO Last.Line
    Out2 = STRIM$(Program(Program.Line))
    IF LEN(Out2) THEN
       Program(Program.Line) = TTRIM$(Out2, True)
    ELSE
       Program(Program.Line) = Nul
    END IF
 NEXT
END SUB

' replaces white spaces with spaces
' var = true; skip white spaces in quotes
FUNCTION TTRIM$(Var$, Var)
 VarX$ = Var$
 Temp = False
 DO
    Temp = Temp + 1
    IF Temp > LEN(VarX$) THEN
       EXIT DO
    END IF
    IF Var THEN
       IF MID$(VarX$, Temp, 1) = Quote THEN
          DO
             Temp = Temp + 1
             IF Temp > LEN(VarX$) THEN
                EXIT DO
             END IF
             IF MID$(VarX$, Temp, 1) = Quote THEN
                EXIT DO
             END IF
          LOOP
       END IF
    END IF
    FOR Blanks = 1 TO LEN(White.Space)
       IF MID$(VarX$, Temp, 1) = MID$(White.Space, Blanks, 1) THEN
          MID$(VarX$, Temp, 1) = " "
       END IF
    NEXT
 LOOP
 Temp = False
 DO
    Temp = Temp + 1
    IF Temp > LEN(VarX$) THEN
       EXIT DO
    END IF
    IF Var THEN
       IF MID$(VarX$, Temp, 1) = Quote THEN
          DO
             Temp = Temp + 1
             IF Temp > LEN(VarX$) THEN
                EXIT DO
             END IF
             IF MID$(VarX$, Temp, 1) = Quote THEN
                EXIT DO
             END IF
          LOOP
       END IF
    END IF
    IF MID$(VarX$, Temp, 2) = "  " THEN
       VarX$ = LEFT$(VarX$, Temp) + MID$(VarX$, Temp + 2)
       Temp = Temp - 1
    END IF
 LOOP
 TTRIM$ = VarX$
END FUNCTION

' strips leading/trailing white spaces from string
FUNCTION STRIM$(Var$)
 VarX$ = Var$
 DO
    Blanks = False
    FOR Count = 1 TO LEN(White.Space)
       IF LEFT$(VarX$, 1) = MID$(White.Space, Count, 1) THEN
          VarX$ = MID$(VarX$, 2)
          Blanks = True
       END IF
    NEXT
    IF Blanks = False THEN
       EXIT DO
    END IF
 LOOP
 DO
    Blanks = False
    FOR Count = 1 TO LEN(White.Space)
       IF RIGHT$(VarX$, 1) = MID$(White.Space, Count, 1) THEN
          VarX$ = LEFT$(VarX$, LEN(VarX$) - 1)
          Blanks = True
       END IF
    NEXT
    IF Blanks = False THEN
       EXIT DO
    END IF
 LOOP
 STRIM$ = VarX$
END FUNCTION

' replaces white spaces with nul,
' var = true; skip white spaces in quotes
FUNCTION XTRIM$(Var$, Var)
 VarX$ = Var$
 Temp = False
 DO
    Temp = Temp + 1
    IF Temp > LEN(VarX$) THEN
       EXIT DO
    END IF
    IF Var THEN
       IF MID$(VarX$, Temp, 1) = CHR$(34) THEN
          DO
             Temp = Temp + 1
             IF Temp > LEN(VarX$) THEN
                EXIT DO
             END IF
             IF MID$(VarX$, Temp, 1) = CHR$(34) THEN
                EXIT DO
             END IF
          LOOP
       END IF
    END IF
    FOR Blanks = 1 TO LEN(White.Space)
       IF MID$(VarX$, Temp, 1) = MID$(White.Space, Blanks, 1) THEN
          VarX$ = LEFT$(VarX$, Temp - 1) + MID$(VarX$, Temp + 1)
          Temp = Temp - 1
          EXIT FOR
       END IF
    NEXT
 LOOP
 XTRIM$ = VarX$
END FUNCTION

' returns a randomized temporary filename.
'    could be generated from date/time.
FUNCTION TempName$
 Var$ = Nul
 FOR Temp = 1 TO 8
    Var$ = Var$ + HEX$(INT(RND*16)) ' 0-9,A-F
 NEXT
 TempName$ = Var$
END FUNCTION

' uses direct keyboard input to get an input line during redirected input.
FUNCTION KeyboardLine$
 Var$ = Nul
 T! = TIMER
 DO
    IF TIMER - T! >= 1! THEN
       X = CSRLIN
       Y = POS(0)
       CALL StatusLine
       COLOR Plain, Black
       LOCATE X, Y, 1
       PRINT "";
       T! = TIMER
    END IF
    IF KeyIS THEN ' check keyboard
       Var2$ = KeyboardChar$ ' get character
       IF LEN(Var2$) = 1 THEN
          SELECT CASE Var2$
          CASE CHR$(9) ' tab
             FOR VarX = 1 TO 8
                PRINT " ";
                LOCATE , , Visible
                Var$ = Var$ + " "
             NEXT
          CASE CHR$(13) ' return
             EXIT DO
          CASE CHR$(8) ' backspace
             IF LEN(Var$) > 0 THEN
                Var$ = LEFT$(Var$, LEN(Var$) - 1)
                GOSUB Back.Space1
             END IF
          CASE CHR$(27) ' escape
             FOR VarX = 1 TO LEN(Var$)
                GOSUB Back.Space1
             NEXT   
             Var$ = Nul
          CASE ELSE ' store input
             IF ASC(Var2$) >= 32 THEN
                PRINT Var2$;
                LOCATE , , Visible
                Var$ = Var$ + Var2$
             END IF
          END SELECT
       END IF
    END IF

    ' check control-break key
    IF BreakIS THEN
       Var$ = Nul
       EXIT DO
    END IF
 LOOP

 ' return input line
 KeyboardLine$ = Var$
 EXIT FUNCTION

Back.Space1:
 IF POS(0) > 1 THEN ' check cursor location
    LOCATE CSRLIN, POS(0) - 1, 0
    PRINT " ";
    LOCATE CSRLIN, POS(0) - 1, Visible
 ELSE ' line-wrap
    LOCATE CSRLIN - 1, ScreenWidth, 0
    PRINT " ";
    LOCATE CSRLIN - 1, ScreenWidth, Visible
 END IF
 RETURN
END FUNCTION

' uses direct keyboard input to get an input line
'   and edit multi-dimensional command history arrays with line-wrapping,
'     where Temp0 (history array type) equals:
'       1=command prompt
'       2=whatis prompt
'       3=debug prompt

FUNCTION KeyboardLine2$
 History.Count2(Temp0)=0 ' reset
 Var$ = Nul
 T! = TIMER
 DO
    IF TIMER - T! >= 1! THEN
       X = CSRLIN
       Y = POS(0)
       CALL StatusLine
       COLOR Plain, Black
       LOCATE X, Y, 1
       PRINT "";
       T! = TIMER
    END IF
    IF KeyIS THEN ' check keyboard
       Var2$ = KeyboardChar$ ' get character
       SELECT CASE LEN(Var2$)
       CASE 1
          SELECT CASE Var2$
          CASE CHR$(3), CHR$(11) ' control-c/control-k
             KeyboardLine2$ = ""
             EXIT FUNCTION
          CASE CHR$(9) ' tab
             IF LEN(Var$) + 8 <= 256 THEN
                FOR VarX = 1 TO 8
                   PRINT " ";
                   LOCATE , , Visible
                   Var$ = Var$ + " "
                NEXT
             END IF
          CASE CHR$(13) ' return
             Gosub Enter.History
             EXIT DO
          CASE CHR$(8) ' backspace
             IF LEN(Var$) > 0 THEN
                Var$ = LEFT$(Var$, LEN(Var$) - 1)
                Gosub Back.Space
             END IF
          CASE CHR$(27) ' escape
             Gosub Erase.Line
          CASE ELSE ' store input
             IF ASC(Var2$) >= 32 THEN
                If Len(Var$) < 256 Then
                   PRINT Var2$;
                   LOCATE , , Visible
                   Var$ = Var$ + Var2$
                END IF
             Endif
          END SELECT
       CASE 2
          SELECT CASE ASC(RIGHT$(Var2$,1))
          Case 77 ' right
             Gosub Get.Count
             ' return blank line at end of history
             IF Var1 > 0 THEN
                If Var1 = History.Count2(Temp0) Then
                    EndOfHistoryFlag(Temp0) = -1
                    GoSub Erase.Line
                    Var$ = ""
                    Var1 = 0
                    Print Var$;
                End If
             End If
             ' move to next history
             If Var1>0 Then
                Select Case History.Count2(Temp0)
                Case 0
                   Gosub Erase.Line
                   History.Count2(Temp0)=Var1
                   Var$=Rtrim$(History2(Temp0,Var1))
                   Print Var$;
                Case 1 To Max.History - 1
                   If History.Count2(Temp0)+1<=Var1 Then
                      Gosub Erase.Line
                      History.Count2(Temp0)=History.Count2(Temp0)+1
                      Var$=Rtrim$(History2(Temp0,History.Count2(Temp0)))
                      Print Var$;
                   Else
                      If History.Count2(Temp0)=Var1 Then
                         Gosub Erase.Line
                         Var$=Rtrim$(History2(Temp0,Var1))
                         Print Var$;
                      Endif
                   Endif
                Case Max.History
                   Gosub Erase.Line
                   Var$=Rtrim$(History2(Temp0,Max.History))
                   Print Var$;
                End Select
             Endif
          Case 75 ' left
             Gosub Get.Count
             If Var1 > 0 Then
                If EndOfHistoryFlag(Temp0) Then
                   EndOfHistoryFlag(Temp0) = 0
                   GoSub Erase.Line
                   History.Count2(Temp0) = Var1
                   Var$ = RTrim$(History2(Temp0, Var1))
                   Print Var$;
                   Var1 = 0
                End If
             End If
             If Var1>0 Then
                Select Case History.Count2(Temp0)
                Case 0
                   Gosub Erase.Line
                   History.Count2(Temp0)=Var1
                   Var$=Rtrim$(History2(Temp0,Var1))
                   Print Var$;
                Case 1
                   Gosub Erase.Line
                   Var$=Rtrim$(History2(Temp0,1))
                   Print Var$;
                Case Else
                   If History.Count2(Temp0)-1>0 Then
                      Gosub Erase.Line
                      History.Count2(Temp0)=History.Count2(Temp0)-1
                      Var$=Rtrim$(History2(Temp0,History.Count2(Temp0)))
                      Print Var$;
                   Endif
                End Select
             Endif
          Case 71 ' home
             Gosub Get.Count
             If Var1>0 Then
                Gosub Erase.Line
                History.Count2(Temp0)=1
                Var$=Rtrim$(History2(Temp0,1))
                Print Var$;
             Endif
          Case 79 ' end
             Gosub Get.Count
             If Var1>0 Then
                Gosub Erase.Line
                History.Count2(Temp0)=Var1
                Var$=Rtrim$(History2(Temp0,Var1))
                Print Var$;
             Endif
          Case 83 ' delete
             Select Case Debug.Mode
             Case 0
                Debug.Mode = -1
             Case -1
                CALL ClearStatusLine
                Color 14
                Print ">";
                Debug.Mode = -2
             Case -2
                Debug.Mode = 0
             End Select
             If Debug.Mode = -2 Then
                ' nul
             Else
                X = CSRLIN
                Y = POS(0)
                CALL StatusLine
                COLOR Plain, Black
                LOCATE X, Y, 1
                PRINT "";
             Endif
          END SELECT
       END SELECT
    END IF

    ' check control-break key
    IF BreakIS THEN
       Var$ = Nul
       EXIT DO
    END IF
 LOOP

 ' return input line
 KeyboardLine2$ = Var$
 EXIT FUNCTION

Enter.History:
 If Len(Var$) Then
    Gosub Get.Count
    ' init
    EndOfHistoryFlag(Temp0) = -1
    If Var1=0 Then
       History.Count2(Temp0)=1
       History2(Temp0,1)=Var$
    Else
       ' add history
       If History.Count2(Temp0)=0 Then
          If Var1<Max.History Then
             History.Count2(Temp0)=Var1+1
             History2(Temp0,Var1+1)=Var$
          Else
             For Var=1 To Max.History - 1
                History2(Temp0,Var)=History2(Temp0,Var+1)
             Next
             History.Count2(Temp0)=Max.History
             History2(Temp0,Max.History)=Var$
          Endif
       Else
          ' pack history
          If Var1=Max.History Then
             For Var=History.Count2(Temp0) To Max.History - 1
                History2(Temp0,Var)=History2(Temp0,Var+1)
             Next
             History.Count2(Temp0)=Max.History
             History2(Temp0,Max.History)=Var$
          Else
             For Var=History.Count2(Temp0) To Var1-1
                History2(Temp0,Var)=History2(Temp0,Var+1)
             Next
             History2(Temp0,Var1)=Var$
          Endif
       Endif
    Endif      
 Endif
 Return

Erase.Line:
 FOR VarX = 1 To LEN(Var$)
    Gosub Back.Space
 NEXT   
 Var$ = Nul
 Return

Back.Space:
 IF POS(0) > 1 THEN ' check cursor location
    LOCATE CSRLIN, POS(0) - 1, 0
    PRINT " ";
    LOCATE CSRLIN, POS(0) - 1, Visible
 ELSE ' line-wrap
    LOCATE CSRLIN - 1, ScreenWidth, 0
    PRINT " ";
    LOCATE CSRLIN - 1, ScreenWidth, Visible
 END IF
 Return

Get.Count:
 Var1=0
 For Var2=Max.History To 1 Step -1
    If Rtrim$(History2(Temp0,Var2))<>Nul Then
       Var1=Var2
       Exit For
    Endif
 Next
 Return
END FUNCTION

' reads character directly from keyboard including extended keys.
FUNCTION KeyboardChar$
 InregsX.AX = &H0000
 CALL InterruptX(&H16, InregsX, OutregsX)
 Key1 = (OutregsX.AX AND &HFF)
 IF Key1 = False THEN ' extended key
    Key2 = (OutregsX.AX AND &HFF00) / 256
    KeyboardChar$ = CHR$(0) + CHR$(Key2)
 ELSE
    KeyboardChar$ = CHR$(Key1)
    IF Key1 = 3 THEN
       ControlBreak = True
    END IF
 END IF
 Flag = BreakIS ' check control-break
END FUNCTION

' checks keyboard buffer
FUNCTION KeyIS
 InregsX.AX = &H0100
 CALL InterruptX(&H16, InregsX, OutregsX)
 IF (OutregsX.Flags AND &H40) = &H40 THEN
    KeyIS = False
 ELSE
    KeyIS = True
 END IF
END FUNCTION

' clears keyboard buffer
FUNCTION ClearBuffer
 DO
    IF KeyIS THEN
       Char$ = KeyboardChar$
    ELSE
       EXIT DO
    END IF
 LOOP
 ClearBuffer = True
END FUNCTION

' releases time slice in windows during tight keyboard input loops.
SUB Release.Time(Var)
 IF Supported.Call = False THEN
    FOR Var1 = 1 TO Var
       InregsX.AX = &H1680
       InregsX.BX = &H0000
       CALL InterruptX(&H2F, InregsX, OutregsX)
       IF (OutregsX.AX AND &HFF) = &H80 THEN
          Supported.Call = True
          EXIT FOR
       END IF
    NEXT
 END IF
END SUB

' checks Control-Break flag
FUNCTION BreakIS
 DEF SEG = &H40 ' set segment
 IF PEEK(&H71) = 128 THEN ' check bit flag
    ControlBreak = True ' store control-break variable
    POKE &H71, &H0 ' reset bit flag
 END IF
 IF InDEFSEG THEN ' check segment flag
    DEF SEG = DEFSEGvalue ' restore currently set segment
 ELSE
    DEF SEG ' restore BASIC segment
 END IF
 IF ControlBreak THEN ' check control-break flag
    BreakIS = True
 ELSE
    BreakIS = False
 END IF
END FUNCTION

' clears Control-Break flag
FUNCTION ClearBreak
 DEF SEG = &H40
 POKE &H71, &H0
 DEF SEG
 ControlBreak = False
 ClearBreak = True
END FUNCTION

' closes last file open,
' resets next free file number
FUNCTION FreeFileNumber
 IF FileNumber THEN
    CLOSE #FileNumber
    FileNumber = False
 END IF
 FileNumber = FreeFile
 FreeFileNumber = True
END FUNCTION

' set prompts
SUB Set.Prompts
 DO
    COLOR White, Black
    PRINT "Edit Prompt Menu:"
    COLOR Yellow, Black
    PRINT "(1)SIC prompt: " + Prompt1
    PRINT "(2)Whatis prompt: " + Prompt2
    PRINT "(3)Debug prompt: " + Prompt3
    DO
       COLOR White, Black
       LOCATE , 1, 1
       PRINT "Enter Prompt(Q To Quit)? ";
       LOCATE , , 1
       Out2 = KeyboardLine$
       PRINT
       If Out2 = Nul Then
          Exit Sub
       Endif
       IF Out2 <> Nul THEN
          EXIT DO
       END IF
    LOOP
    If Ucase$(Out2)="Q" Then
       Exit Do
    Endif
    COLOR White, Black
    SELECT CASE INT(VAL(Out2))
    CASE 1
       PRINT "SIC Prompt:";
       Out2 = KeyboardLine$
       PRINT
       Prompt1 = Out2
    CASE 2
       PRINT "Whatis Prompt:";
       Out2 = KeyboardLine$
       PRINT
       Prompt2 = Out2
    CASE 3
       PRINT "Debug Prompt:";
       Out2 = KeyboardLine$
       PRINT
       Prompt3 = Out2
    END SELECT  
 LOOP
END SUB

' function to match case-sensitive substring with ?, * characters in substring
SUB InstrSUB1 (Var, Var1$, Var2$, VarQ)

' store case-sensitive string match variables
IF VarQ THEN
   S2$ = Var1$
   S3$ = Var2$
ELSE
   S2$ = LCASE$(Var1$)
   S3$ = LCASE$(Var2$)
END IF

' check default instr
IF INSTR(S2$, "*") = 0 THEN
   IF INSTR(S2$, "?") = 0 THEN
      Var = INSTR(S3$, S2$)
      EXIT SUB
   END IF
END IF

Var = -1 ' assume match

' see if S2$ matches in S3$ with substrings
FOR S3 = 1 TO LEN(S3$)
   S1$ = MID$(S3$, S3)
   P1 = 1 ' pointer to S1$
   P2 = 1 ' pointer to S2$
   DO
      ' check match
      IF P2 > LEN(S2$) THEN
         EXIT SUB
      END IF

      ' check character in S2$ at P2
      V$ = MID$(S2$, P2, 1)
      SELECT CASE V$
         CASE "*" ' global character
            ' scan to next char
            IF P2 > LEN(S2$) THEN
               EXIT DO
            END IF
            S4$ = MID$(S2$, P2 + 1, 1)
            SELECT CASE S4$
               CASE "*", "?"
                  P2 = P2 + 1
               CASE ELSE
                  DO
                     IF MID$(S1$, P1, 1) = S4$ THEN
                        EXIT DO
                     END IF
                     IF P1 >= LEN(S1$) THEN
                        EXIT DO
                     END IF
                     P1 = P1 + 1
                  LOOP
                  P2 = P2 + 1
            END SELECT
         CASE "?" ' wildcard character
            P1 = P1 + 1
            P2 = P2 + 1
         CASE ELSE ' ascii character
            IF MID$(S1$, P1, 1) <> V$ THEN ' no match
               EXIT DO
            END IF
            P1 = P1 + 1
            P2 = P2 + 1
      END SELECT
   LOOP
NEXT
Var = 0 ' no match
END SUB

SUB KillTemp(VarX)
 ON LOCAL ERROR GOTO Error1
 IF VarX = 0 THEN
    KILL "C:\TEMP"
 ELSE
    KILL "C:\TEMP" + LTRIM$(STR$(VarX))
 END IF
 EXIT SUB
Error1:
 EXIT SUB
END SUB

SUB MakeTemp(VarX)
 ON LOCAL ERROR GOTO Error2
 IF VarX = 0 THEN
    MKDIR "C:\TEMP"
 ELSE
    MKDIR "C:\TEMP" + LTRIM$(STR$(VarX))
 END IF
 EXIT SUB
Error2:
 EXIT SUB
END SUB

' detects max rows
Sub Read.Rows(Var)
 On Local Error Goto Error.Trap62x
 Temp1=50
 Locate 50,1,0
 Var=50
 If PageLength=0 Then PageLength=48
 Exit Sub
Next.Trap1:
 Temp1=43
 Locate 43,1,0
 Var=43
 If PageLength=0 Then PageLength=41
 Exit Sub
Next.Trap2:
 Temp1=25
 Locate 25,1,0
 Var=25
 If PageLength=0 Then PageLength=23
 Exit Sub
Next.Trap3:
 Var=12
 If PageLength=0 Then PageLength=10
Error.Resume62x:
 Exit Sub
Error.Trap62x:
 If Temp1=50 Then
    Resume Next.Trap1
 Endif
 If Temp1=43 Then
    Resume Next.Trap2
 Endif
 If Temp1=25 Then
    Resume Next.Trap3
 Endif
 Resume Error.Resume62x
End Sub

SUB ClearStatusLine
 If StdinEnabled = 0 Then
    CLS
    CALL Save.Screen2(-1)
    IF TempX = 0 THEN TempX = 1
    LOCATE TempX, 1, 1
    PRINT
    EXIT SUB
 Endif
 LOCATE Max.Row, 1, 1
 COLOR Plain, Black
 PRINT SPACE$(80);
 LOCATE Max.Row - 1, 1, 1
END SUB

SUB Save.Screen2 (Var%)
 ON LOCAL ERROR GOTO Error.Trap62
 STATIC Temp.ArrayY() AS INTEGER
 STATIC Temp.ArrayZ() AS INTEGER
 IF Var% = False THEN
    VarX3 = (Max.Row - 1) * 80 + 80
    REDIM Temp.ArrayY(VarX3) AS INTEGER
    REDIM Temp.ArrayZ(VarX3) AS INTEGER
    FOR Var1 = 1 TO Max.Row
       FOR Var2 = 1 TO 80
          TempZ1 = SCREEN(Var1, Var2)
          TempZ2 = SCREEN(Var1, Var2, 1)
          VarX2 = (Var1 - 1) * 80 + Var2
          Temp.ArrayY(VarX2) = TempZ1' char
          Temp.ArrayZ(VarX2) = TempZ2' color
       NEXT
    NEXT
 ELSE
    CLS
    FOR VarX = Max.Row TO 1 STEP -1
       FOR VarY = 1 TO 80
          VarX2 = (VarX - 1) * 80 + VarY
          IF Temp.ArrayY(VarX2) <> 32 THEN
             GOTO Next.Line
          END IF
       NEXT
    NEXT
Next.Line:
    TempX = VarX
    FOR Var1 = 1 TO VarX ' Max.Row
       FOR Var2 = 1 TO 80
          VarX2 = (Var1 - 1) * 80 + Var2
          VarB = INT(Temp.ArrayZ(VarX2) / 16)
          VarF = Temp.ArrayZ(VarX2) MOD 16
          TempZ1 = Temp.ArrayY(VarX2)
          LOCATE Var1, Var2, 0
          COLOR VarF, VarB
          PRINT CHR$(TempZ1);
       NEXT
    NEXT
 END IF
Error.Resume62:
 EXIT SUB
Error.Trap62:
 RESUME Error.Resume62
END SUB

SUB StatusLine
 If StdinEnabled Then
    Exit Sub
 Endif
 If Debug.Mode = -2 Then
    Exit Sub
 Endif
 S$ = "v" + Version$ + " r" + Release$ + " " + Format$(Now, "mm-dd-yyyy") + " " + Format$(Now, "hh:mm:ss")
 If Debug.Mode Then
    Var2# = FRE(-1) + FRE("A") ' far+near segments
    Var2$ = FORMAT$(Var2#, "#,##0.00;;")
    S$ = "- Free RAM: " + Var2$ + " bytes."
 END IF
 LOCATE Max.Row, 1, 1
 COLOR White, Blue
 S$ = "SIC utility " + S$
 S$ = LEFT$(S$, 80)
 S$ = S$ + SPACE$(80 - LEN(S$))
 PRINT S$;
 COLOR Yellow, Black
 LOCATE Max.Row-1,1,1
END SUB

' declares all 24 Sic commands
Command.Data:
 DATA "ANALYZE","FILES","HELP","INDENT","KILL", "LIST","LOAD","NEW","PRINT","QUIT"
 DATA "RENUMBER","RUN","SAVE","WHATIS","CONTINUE","DEBUG","SHELL","SEARCH","SET"
 DATA "TIME","DATE","CLOCK","DIRS","DRIVES"

' declares indent structure types
Indent.Data:
 ' 1=indent right
 DATA "SELECTIF CASE", 1
 DATA "DO WHILE", 1
 DATA "DO", 1
 DATA "IF", 1
 DATA "SELECT CASE", 1
 DATA "DO UNTIL", 1
 DATA "FORIF", 1
 DATA "FOR", 1
 DATA "WHILE", 1
DATA "LOOPIF", 1
 ' -1=indent left
 DATA "ENDIF", -1
 DATA "END IF", -1
 DATA "NEXTIF", -1
 DATA "NEXT", -1
 DATA "LOOP UNTIL", -1
 DATA "LOOP WHILE", -1
 DATA "END SELECT", -1
 DATA "WEND", -1
 DATA "END LOOPIF", -1
 DATA "LOOP", -1
 DATA "END SELECTIF", -1
 ' -2=indent left, then right
 DATA "ELSE", -2
 DATA "CASE", -2
 DATA "EOF", 0

' declares analyze functions types and pairs
Analyze.Data1:
 DATA "IF", "ENDIF"
 DATA "DO", "LOOP"
 DATA "FOR", "NEXT"
 DATA "WHILE", "WEND"
 DATA "FORIF", "NEXTIF"
 DATA "LOOPIF", "END LOOPIF"
 DATA "SELECT CASE", "END SELECT"
 DATA "SELECTIF CASE", "END SELECTIF"

Analyze.Data2:
 DATA "IF", "ENDIF", "ELSE", "ELSEIF"
 DATA "SELECT CASE", "END SELECT", "CASE ELSE", "CASE"
 DATA "SELECTIF CASE", "END SELECTIF", "CASEIF ELSE", "CASEIF"

Analyze.Data3:
 DATA "DO", "LOOP", "EXIT DO", "CONTINUE DO"
 DATA "FOR", "NEXT", "EXIT FOR", "CONTINUE FOR"
 DATA "WHILE", "WEND", "EXIT WHILE", "CONTINUE WHILE"
 DATA "FORIF","NEXTIF", "EXIT FORIF", "CONTINUE FORIF"
 DATA "LOOPIF", "END LOOPIF", "EXIT LOOPIF", "CONTINUE LOOPIF"

 ' declares all Sic statements on left side of equation,
 ' each statement listed contains its own subroutine.
 ' listed in rows of ten.
Program.Data:
 DATA "'","ENDIF","END IF","STOP","REM","MID$","LEFT$","RIGHT$","PRINT #","DPRINT"
 DATA "LPRINT USING","SPRINT","UPRINT","INPUT;","FORIF","FOR","NEXTIF","NEXT","CONTINUE FORIF","CONTINUE FOR"
 DATA "EXIT FORIF","EXIT FOR","DO UNTIL","LOOP WHILE","EXIT DO","CONTINUE DO","GOTO","GOSUB","RETURN","DO WHILE"
 DATA "DO","OFF","IF","ELSEIF","CASEIF ELSE","CASEIF","SELECT CASE","END SELECTIF","BEEP","SOUND"
 DATA "COLOR","LOCATE","CLS","SCREEN","WIDTH","WRITE #","LINE INPUT;","LINE INPUT #","INPUT #","WEND"
 DATA "WHILE","CONTINUE WHILE","EXIT WHILE","ELSE","LOOP UNTIL","LOOPIF","END LOOPIF","EXIT LOOPIF","LOOP","RANDOMIZE"
 DATA "POKE","INT86","DEFSEG","ABSOLUTE","OUT","WAIT","SLEEP","PAUSE","SELECTIF CASE","END SELECT"
 DATA "CASE ELSE","CASE","CONTINUE LOOPIF","END","CLEAR","SYSTEM","SWAP","ERROR","ON ERROR GOTO","ON ERROR RESUME PREVIOUS"
 DATA "ON ERROR RESUME SAME","ON ERROR RESUME NEXT","ON ERROR STOP","RESUME PREVIOUS","RESUME SAME","RESUME NEXT","RESUME","ON","DATE$","TIME$"
 DATA "CHDRIVE","CD","CHDIR","MD","MKDIR","RD","RMDIR","KILL","DELETE","RENAME"
 DATA "NAME","SHELL","CHAIN","LET","CLOSE #","OPEN #","FIELD #","WRITE","PRINT USING","INPUT"
 DATA "LINE INPUT","LSET #","RSET #","PUT #","GET #","READ #","DATA","READ","RESTORE","CIRCLE STEP"
 DATA "LINE STEP","PSET STEP","PRESET STEP","PAINT STEP","DRAW","PLAY","GET STEP","PUT STEP","BSAVE","BLOAD"
 DATA "VIEW SCREEN","VIEW","WINDOW SCREEN","WINDOW","CIRCLE","LINE","CLOSE","DEF FN","ENVIRON","PAINT"
 DATA "PSET","PRESET","GET","PUT","DIM","COMMON","DECLARE","LOCK #","UNLOCK #","PRINT","LPRINT"

 REM  "This program is public domain software written by:"
 DATA "Author: Erik Jon Oredson"
 DATA "Email: eoredson@gmail.com"
 DATA "Url: www.filegate.net"
 REM  "Note: Cookies are delicious. So are cats."

 ' -end of program-
